from yankee.json.schema import fields as f
from yankee.json.schema import Schema


class GlobalDossierPublicationSchema(Schema):
    pub_country = f.Str()
    pub_num = f.Str()
    kind_code = f.Str()
    pub_date = f.Date("pubDateStr")


class GlobalDossierPriorityClaimSchema(Schema):
    country = f.Str()
    doc_number = f.Str()
    kind_code = f.Str()


class GlobalDossierDocumentNumberSchema(Schema):
    country = f.Str()
    doc_number = f.Str()
    format = f.Str()
    date = f.Date()
    kind_code = f.Str()


class GlobalDossierApplicationSchema(Schema):
    app_num = f.Str()
    app_date = f.Date("appDateStr")
    country_code = f.Str()
    kind_code = f.Str()
    doc_num = GlobalDossierDocumentNumberSchema("docNum")
    title = f.Str()
    applicant_names = f.Str()
    ip_5 = f.Bool("ip5")
    priority_claim_list = f.List(GlobalDossierPriorityClaimSchema)
    pub_list = f.List(GlobalDossierPublicationSchema)


class GlobalDossierSchema(Schema):
    country = f.Str()
    internal = f.Bool(true_value="true")
    corr_app_num = f.Str()
    id = f.Str()
    type = f.Str()
    applications = f.List(GlobalDossierApplicationSchema, "list")


class DocumentSchema(Schema):
    doc_number = f.Str()
    country = f.Str()
    doc_code = f.Str()
    doc_desc = f.Str()
    doc_id = f.Str()
    date = f.Date("legalDateStr")
    doc_format = f.Str()
    pages = f.Int("numberOfPages")
    doc_code_desc = f.Str()
    doc_group_code = f.Str()
    shareable = f.Bool()


class DocumentListSchema(Schema):
    title = f.Str("title")
    doc_number = f.Str()
    country = f.Str()
    message = f.Str()
    applicant_names = f.List(f.Str())
    office_action_count = f.Int("oaIndCount")
    docs = f.List(DocumentSchema)
    office_action_docs = f.List(DocumentSchema)

    def pre_load(self, obj):
        update_dict = {
            "country": obj["country"],
            "docNumber": obj["docNumber"],
        }
        for doc in obj["docs"]:
            doc.update(update_dict)
        for doc in obj["officeActionDocs"]:
            doc.update(update_dict)
        return obj



from dataclasses import dataclass
from dataclasses import field
from typing import List

from patent_client.epo.ops.util import InpadocModel
from patent_client.util import Model
from patent_client.util.claims.model import Claim


@dataclass
class FTDocumentId(Model):
    country: str = None
    doc_number: str = None
    kind: str = None

    def __str__(self):
        return f"{self.country}{self.doc_number}{self.kind}"


@dataclass
class Claims(Model):
    __manager__ = "patent_client.epo.ops.published.manager.ClaimsManager"
    document_id: FTDocumentId
    claims: List[Claim] = field(default_factory=list)
    claim_text: str = None

    @property
    def docdb_number(self) -> str:
        return str(self.document_id)

    def __repr__(self):
        return f"Claims(document_id={str(self.document_id)})"


@dataclass
class Description(InpadocModel):
    __manager__ = "patent_client.epo.ops.published.manager.DescriptionManager"
    document_id: FTDocumentId = None
    description: str = None

    @property
    def docdb_number(self) -> str:
        return str(self.document_id)



import pytest

from .manager import GlobalDossierDocument
from .model import Document
from .model import DocumentList
from .model import GlobalDossier
from .model import GlobalDossierApplication


class TestGlobalDossier:
    def test_us_lookup(self):
        result = GlobalDossier.objects.get("16123456")
        assert result.id == "16123456"

    def test_us_lookup_docs(self):
        result = GlobalDossier.objects.get("16123456").applications[0].document_list
        assert result.office_action_count == 2

    def test_us_download_docs(self, tmpdir):
        result = GlobalDossier.objects.get("16123456").applications[0].documents[0].download(tmpdir)
        assert result.exists()

    def test_us_app_lookup(self):
        result = GlobalDossierApplication.objects.get("16123456")
        assert result.app_num == "16123456"
        assert isinstance(result, GlobalDossierApplication)

    def test_us_app_lookup_docs(self):
        result = GlobalDossierApplication.objects.get("16123456")
        assert isinstance(result.document_list, DocumentList)
        assert isinstance(result.documents, list)
        assert isinstance(result.documents[0], Document)
        assert isinstance(result.office_actions, list)
        assert isinstance(result.office_actions[0], Document)

    def test_links(self):

        app = GlobalDossierApplication.objects.get("16123456")
        assert (
            app.us_application.patent_title
            == "LEARNING ASSISTANCE DEVICE, METHOD OF OPERATING LEARNING ASSISTANCE DEVICE, LEARNING ASSISTANCE PROGRAM, LEARNING ASSISTANCE SYSTEM, AND TERMINAL DEVICE"
        )
        # Broken links to Public Patent Search
        # assert (
        #    GlobalDossierApplication.objects.get("16123456").us_publication.patent_title
        #    == "Learning assistance device, method of operating learning assistance device, learning assistance program, learning assistance system, and terminal device"
        # )
        # assert (
        #    GlobalDossierApplication.objects.get("16123456").us_patent.patent_title
        #    == "Learning assistance device, method of operating learning assistance device, learning assistance program, learning assistance system, and terminal device"
        # )
        assert app.us_assignments.first().id == "46816-108"
        with pytest.raises(ValueError):
            GlobalDossierApplication.objects.get(publication="EP1000000").us_application
        # Broken links to Public Patent Search
        # with pytest.raises(ValueError):
        #    GlobalDossierApplication.objects.get(publication="EP1000000").us_publication
        # with pytest.raises(ValueError):
        #    GlobalDossierApplication.objects.get(publication="EP1000000").us_patent
        with pytest.raises(ValueError):
            GlobalDossierApplication.objects.get(publication="EP1000000").us_assignments

    def test_raises_errors_on_db_methods(self):
        with pytest.raises(NotImplementedError):
            GlobalDossier.objects.filter("something")
        with pytest.raises(NotImplementedError):
            GlobalDossier.objects.order_by("something")
        with pytest.raises(NotImplementedError):
            GlobalDossier.objects.limit("something")
        with pytest.raises(NotImplementedError):
            GlobalDossier.objects.offset("something")
        with pytest.raises(NotImplementedError):
            GlobalDossierDocument().filter("something")
        with pytest.raises(NotImplementedError):
            GlobalDossierDocument().order_by("something")
        with pytest.raises(NotImplementedError):
            GlobalDossierDocument().limit("something")
        with pytest.raises(NotImplementedError):
            GlobalDossierDocument().offset("something")

    def test_issue_99(self):
        app = GlobalDossierApplication.objects.get("16740760", type="application", office="US")
        assert app.app_num == "16740760"
        app = GlobalDossierApplication.objects.get("17193105")
        assert app.app_num == "17193105"


class TestGlobalDossierAsync:
    @pytest.mark.asyncio
    async def test_us_lookup(self):
        result = await GlobalDossier.objects.aget("16123456")
        assert result.id == "16123456"

    @pytest.mark.asyncio
    async def test_us_lookup_docs(self):
        dossier = await GlobalDossier.objects.aget("16123456")
        result = dossier.applications[0].document_list
        assert result.office_action_count == 2

    @pytest.mark.asyncio
    async def test_us_download_docs(self, tmpdir):
        dossier = await GlobalDossier.objects.aget("16123456")
        result = dossier.applications[0].documents[0].download(tmpdir)
        assert result.exists()

    @pytest.mark.asyncio
    async def test_us_app_lookup(self):
        result = await GlobalDossierApplication.objects.aget("16123456")
        assert result.app_num == "16123456"
        assert isinstance(result, GlobalDossierApplication)

    @pytest.mark.asyncio
    async def test_us_app_lookup_docs(self):
        result = await GlobalDossierApplication.objects.aget("16123456")
        assert isinstance(result.document_list, DocumentList)
        assert isinstance(result.documents, list)
        assert isinstance(result.documents[0], Document)
        assert isinstance(result.office_actions, list)
        assert isinstance(result.office_actions[0], Document)

    @pytest.mark.asyncio
    async def test_links(self):
        app = await GlobalDossierApplication.objects.aget("16123456")
        assert (
            app.us_application.patent_title
            == "LEARNING ASSISTANCE DEVICE, METHOD OF OPERATING LEARNING ASSISTANCE DEVICE, LEARNING ASSISTANCE PROGRAM, LEARNING ASSISTANCE SYSTEM, AND TERMINAL DEVICE"
        )
        # Broken links to Public Patent Search
        # assert (
        #    GlobalDossierApplication.objects.get("16123456").us_publication.patent_title
        #    == "Learning assistance device, method of operating learning assistance device, learning assistance program, learning assistance system, and terminal device"
        # )
        # assert (
        #    GlobalDossierApplication.objects.get("16123456").us_patent.patent_title
        #    == "Learning assistance device, method of operating learning assistance device, learning assistance program, learning assistance system, and terminal device"
        # )
        assert app.us_assignments.first().id == "46816-108"
        with pytest.raises(ValueError):
            GlobalDossierApplication.objects.get(publication="EP1000000").us_application
        # Broken links to Public Patent Search
        # with pytest.raises(ValueError):
        #    GlobalDossierApplication.objects.get(publication="EP1000000").us_publication
        # with pytest.raises(ValueError):
        #    GlobalDossierApplication.objects.get(publication="EP1000000").us_patent
        with pytest.raises(ValueError):
            GlobalDossierApplication.objects.get(publication="EP1000000").us_assignments

    @pytest.mark.asyncio
    async def test_issue_99(self):
        app = await GlobalDossierApplication.objects.aget("16740760", type="application", office="US")
        assert app.app_num == "16740760"
        app = await GlobalDossierApplication.objects.aget("17193105")
        assert app.app_num == "17193105"



import datetime
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path

from patent_client.util.asyncio_util import run_sync
from patent_client.util.base.model import Model
from patent_client.util.base.related import get_model
from yankee.data import ListCollection

from .api import GlobalDossierAsyncApi

global_dossier_async_api = GlobalDossierAsyncApi()


@dataclass
class GlobalDossierPublication(Model):
    pub_country: "str" = None
    pub_num: "str" = None
    kind_code: "str" = None
    pub_date: "datetime.date" = None


@dataclass
class GlobalDossierPriorityClaim(Model):
    country: "str" = None
    doc_number: "str" = None
    kind_code: "str" = None


@dataclass
class GlobalDossierDocumentNumber(Model):
    country: "str" = None
    doc_number: "str" = None
    format: "str" = None
    date: "datetime.date" = None
    kind_code: "str" = None


@dataclass
class GlobalDossierApplication(Model):
    __manager__ = "patent_client.uspto.global_dossier.manager.GlobalDossierApplicationManager"
    app_num: "str" = None
    app_date: "datetime.date" = None
    country_code: "str" = None
    kind_code: "str" = None
    doc_num: "GlobalDossierDocumentNumber" = None
    title: "str" = None
    applicant_names: "str" = None
    ip_5: "bool" = None
    priority_claim_list: "ListCollection[GlobalDossierPriorityClaim]" = field(default_factory=ListCollection)
    pub_list: "ListCollection[GlobalDossierPublication]" = field(default_factory=ListCollection)

    def __repr__(self):
        return f"GlobalDossierApplication(app_num={self.app_num}, country_code={self.country_code})"

    @property
    def document_list(self):
        return get_model("patent_client.uspto.global_dossier.model.DocumentList").objects.get(
            self.country_code, self.app_num, self.kind_code
        )

    @property
    def documents(self):
        return (
            get_model("patent_client.uspto.global_dossier.model.DocumentList")
            .objects.get(self.country_code, self.app_num, self.kind_code)
            .docs
        )

    @property
    def office_actions(self):
        return (
            get_model("patent_client.uspto.global_dossier.model.DocumentList")
            .objects.get(self.country_code, self.app_num, self.kind_code)
            .office_action_docs
        )

    @property
    def us_application(self):
        if self.country_code != "US":
            raise ValueError(f"Global Dossier Application is not a US Application! {self}")
        return get_model("patent_client.uspto.peds.model.USApplication").objects.get(self.app_num)

    @property
    def us_publication(self):
        if self.country_code != "US":
            raise ValueError(f"Global Dossier Application is not a US Application! {self}")
        pub = list(p for p in self.pub_list if p.kind_code == "A1")
        if len(pub) > 1:
            raise ValueError("More than one US publication for application!")
        return get_model("patent_client.uspto.public_search.model.PublishedApplication").objects.get(pub[0].pub_num)

    @property
    def us_patent(self):
        if self.country_code != "US":
            raise ValueError(f"Global Dossier Application is not a US Application! {self}")
        pat = list(p for p in self.pub_list if p.kind_code in ("A", "B1", "B2"))
        if len(pat) > 1:
            raise ValueError("More than one US patent for application!")
        return get_model("patent_client.uspto.public_search.model.Patent").objects.get(pat[0].pub_num)

    @property
    def us_assignments(self):
        if self.country_code != "US":
            raise ValueError(f"Global Dossier Application is not a US Application! {self}")
        return get_model("patent_client.uspto.assignment.model.Assignment").objects.filter(appl_id=self.app_num)


@dataclass
class GlobalDossier(Model):
    __manager__ = "patent_client.uspto.global_dossier.manager.GlobalDossierManager"
    country: "str" = None
    internal: "bool" = None
    corr_app_num: "str" = None
    id: "str" = None
    type: "str" = None
    applications: "ListCollection[GlobalDossierApplication]" = field(default_factory=ListCollection)

    def __repr__(self):
        return f"GlobalDossier(id={self.id}, type={self.type}, country={self.country})"


@dataclass
class Document(Model):
    doc_number: "str" = None
    country: "str" = None
    doc_code: "str" = None
    doc_desc: "str" = None
    doc_id: "str" = None
    date: "datetime.date" = None
    doc_format: "str" = None
    pages: "int" = None
    doc_code_desc: "str" = None
    doc_group_code: "str" = None
    shareable: "bool" = None

    def download(self, filename="", path="."):
        return run_sync(self.adownload(filename, path))

    async def adownload(self, filename="", path="."):
        out_path = Path(path).expanduser() / (
            f'{self.date.isoformat()} - {self.doc_desc.replace("/", "_")}.pdf' if not filename else filename
        )
        return await global_dossier_async_api.get_document(
            self.country, self.doc_number, self.doc_id, out_path=out_path
        )


@dataclass
class DocumentList(Model):
    __manager__ = "patent_client.uspto.global_dossier.manager.GlobalDossierDocument"
    title: "str" = None
    doc_number: "str" = None
    country: "str" = None
    message: "str" = None
    applicant_names: "ListCollection[str]" = field(default_factory=ListCollection)
    office_action_count: "int" = None
    docs: "ListCollection[Document]" = field(default_factory=ListCollection)
    office_action_docs: "ListCollection[Document]" = field(default_factory=ListCollection)



import re

from patent_client.epo.ops.util import Schema
from yankee.xml import fields as f

doc_number_re = re.compile(r"published-data/images/(?P<country>[^/]+)/(?P<doc_number>[^/]+)/(?P<kind_code>[^/]+)/")


def get_doc_number(string):
    data = doc_number_re.search(string).groupdict()
    return f"{data['country']}{data['doc_number']}{data['kind_code']}"


class DocDbSchema(f.Combine):
    country = f.Str('.//epo:document-id[@document-id-type="docdb"]/epo:country')
    doc_number = f.Str('.//epo:document-id[@document-id-type="docdb"]/epo:doc-number')
    kind = f.Str('.//epo:document-id[@document-id-type="docdb"]/epo:kind')

    def combine_func(self, obj):
        return f"{obj.country}{obj.doc_number}{obj.kind}"


class SectionSchema(Schema):
    name = f.Str("./@name")
    start_page = f.Int("./@start-page")


class ImageDocumentSchema(Schema):
    num_pages = f.Int("./@number-of-pages")
    description = f.Str("./@desc")
    link = f.Str("./@link")
    formats = f.List(f.Str(), "./ops:document-format-options/ops:document-format")
    sections = f.List(SectionSchema, "./ops:document-section")
    doc_number = f.Str("./@link", formatter=get_doc_number)


class ImagesSchema(Schema):
    # search_reference = DocumentIdSchema(".//ops:document-inquiry/ops:publication-reference")
    publication_number = DocDbSchema(".//ops:inquiry-result/epo:publication-reference")
    full_document = ImageDocumentSchema('.//ops:document-instance[@desc="FullDocument"]')
    drawing = ImageDocumentSchema('.//ops:document-instance[@desc="Drawing"]')
    first_page = ImageDocumentSchema('.//ops:document-instance[@desc="FirstPageClipping"]')



from .session import session


class GlobalDossierAsyncApi:
    async def get_file(self, doc_number, type_code="application", office_code="US"):
        url = f"https://gd-api2.uspto.gov/patent-family/svc/family/{type_code}/{office_code}/{doc_number}"
        pre_flight = await session.options(url)
        response = await session.get(url)
        response.raise_for_status()
        return response.json()

    async def get_doc_list(self, country, doc_number, kind_code):
        url = f"https://gd-api2.uspto.gov/doc-list/svc/doclist/{country}/{doc_number}/{kind_code}"
        response = await session.get(url)
        response.raise_for_status()
        return response.json()

    async def get_document(self, country, doc_number, document_id, out_path):
        if out_path.exists():
            return out_path

        url = f"https://gd-api2.uspto.gov/doc-content/svc/doccontent/{country}/{doc_number}/{document_id}/1/PDF"
        async with session.stream(url, stream=True) as response:
            response.raise_for_status()
            with out_path.open("wb") as f:
                async for chunk in response.aiter_bytes(chunk_size=8192):
                    if chunk:
                        f.write(chunk)
        return out_path



from patent_client.epo.ops.number_service.schema import DocumentIdSchema
from patent_client.epo.ops.util import Schema
from yankee.util import clean_whitespace
from yankee.xml import fields as f


class CpcClassificationSchema(f.Combine):
    section = f.Str("./epo:section")
    klass = f.Str("./epo:class")
    subclass = f.Str("./epo:subclass")
    main_group = f.Str("./epo:main-group")
    subgroup = f.Str("./epo:subgroup")
    value = f.Str("./epo:classification-value")
    generating_office = f.Str("./epo:generating-office")

    def combine_func(self, obj):
        return f"{obj.section}{obj.klass}{obj.subclass} {obj.main_group}/{obj.subgroup}"


class CitationSchema(Schema):
    cited_phase = f.Str("./@cited-phase")
    cited_by = f.Str("./@cited-by")
    epodoc = DocumentIdSchema('.//epo:document-id[@document-id-type="epodoc"]')
    docdb = DocumentIdSchema('.//epo:document-id[@document-id-type="docdb"]')
    original = DocumentIdSchema('.//epo:document-id[@document-id-type="original"]')


class TitleSchema(Schema):
    lang = f.Str("./@lang")
    text = f.Str()


class DocDbNumberField(f.Combine):
    country = f.Str(".//epo:country")
    number = f.Str(".//epo:doc-number")
    kind = f.Str(".//epo:kind")

    def combine_func(self, obj):
        if not obj:
            return None
        return f"{obj.get('country', '')}{obj.number}{obj.get('kind', '')}"


class InpadocBiblioSchema(Schema):
    country = f.Str("./@country")
    doc_number = f.Str("./@doc-number")
    kind = f.Str("./@kind")
    family_id = f.Str("./@family-id")
    publication_number = DocDbNumberField('.//epo:publication-reference/epo:document-id[@document-id-type="docdb"]')
    publication_reference_docdb = DocumentIdSchema(
        './/epo:publication-reference/epo:document-id[@document-id-type="docdb"]'
    )
    publication_reference_epodoc = DocumentIdSchema(
        './/epo:publication-reference/epo:document-id[@document-id-type="epodoc"]'
    )
    publication_reference_original = DocumentIdSchema(
        './/epo:publication-reference/epo:document-id[@document-id-type="original"]'
    )
    application_number = DocDbNumberField('.//epo:application-reference/epo:document-id[@document-id-type="docdb"]')
    application_reference_docdb = DocumentIdSchema(
        './/epo:application-reference/epo:document-id[@document-id-type="docdb"]'
    )
    application_reference_epodoc = DocumentIdSchema(
        './/epo:application-reference/epo:document-id[@document-id-type="epodoc"]'
    )
    application_reference_original = DocumentIdSchema(
        './/epo:application-reference/epo:document-id[@document-id-type="original"]'
    )
    intl_class = f.List(
        f.Str(formatter=clean_whitespace),
        "./epo:bibliographic-data/epo:classifications-ipcr/epo:classification-ipcr",
    )
    cpc_class = f.List(
        CpcClassificationSchema,
        './/epo:patent-classifications/epo:patent-classification/epo:classification-scheme[@scheme="CPCI"]/ancestor::epo:patent-classification',
    )
    us_class = f.List(
        f.Str(),
        './/epo:classification-scheme[@scheme="UC"]/following-sibling::epo:classification-symbol',
    )
    priority_claims = f.List(DocumentIdSchema, ".//epo:priority-claim/epo:document-id")
    title = f.Str('.//epo:invention-title[@lang="en"]')
    titles = f.List(TitleSchema, ".//epo:invention-title")
    abstract = f.Str('.//epo:abstract[@lang="en"]')
    citations = f.List(CitationSchema, ".//epo:citation")
    applicants_epodoc = f.List(f.Str(), './/epo:applicant[@data-format="epodoc"]//epo:name')
    applicants_original = f.List(f.Str(), './/epo:applicant[@data-format="original"]//epo:name')
    inventors_epodoc = f.List(f.Str(), './/epo:inventor[@data-format="epodoc"]')
    inventors_original = f.List(f.Str(), './/epo:inventor[@data-format="original"]')


class BiblioResultSchema(Schema):
    documents = f.List(InpadocBiblioSchema, ".//epo:exchange-document")



import re

from patent_client.epo.ops.util import Schema
from patent_client.util.claims.parser import ClaimsParser
from yankee.xml import fields as f


class FTDocumentIdSchema(Schema):
    country = f.Str("./ft:country")
    doc_number = f.Str("./ft:doc-number")
    kind = f.Str("./ft:kind")


class ClaimsSchema(Schema):
    document_id = FTDocumentIdSchema(".//ft:document-id")
    claims = f.Str(".//ft:claims", formatter=ClaimsParser().parse)
    claim_text = f.Str(".//ft:claims", formatter=lambda text: text)


class DescriptionField(f.Field):
    format_re = re.compile(r"\s+\n+\s+")

    def post_load(self, obj):
        text = "\n\n".join(p.strip() for p in obj.itertext()).strip()
        return self.format_re.sub("\n\n", text)


class DescriptionSchema(Schema):
    document_id = FTDocumentIdSchema(".//ft:document-id")
    description = DescriptionField(".//ft:description")



from .base.manager import Manager
from .base.manager import ModelType
from .base.model import Model
from .base.related import get_manager
from .base.related import one_to_many
from .base.related import one_to_one


class DefaultDict(dict):
    def __init__(self, *args, default=None, **kwargs):
        self.default = default
        super().__init__(*args, **kwargs)

    def __missing__(self, key):
        return self.default


__all__ = [
    "DefaultDict",
    "Manager",
    "ModelType",
    "Model",
    "get_manager",
    "one_to_many",
    "one_to_one",
]



import unittest
from datetime import date
from datetime import datetime
from datetime import time
from datetime import timedelta

from .date_parse import get_fixed_timezone
from .date_parse import parse_date
from .date_parse import parse_datetime
from .date_parse import parse_duration
from .date_parse import parse_time


class DateParseTests(unittest.TestCase):
    def test_parse_date(self):
        # Valid inputs
        self.assertEqual(parse_date("2012-04-23"), date(2012, 4, 23))
        self.assertEqual(parse_date("2012-4-9"), date(2012, 4, 9))
        # Invalid inputs
        self.assertIsNone(parse_date("2012423"))
        with self.assertRaises(ValueError):
            parse_date("2012-04-56")

    def test_parse_time(self):
        # Valid inputs
        self.assertEqual(parse_time("09:15:00"), time(9, 15))
        self.assertEqual(parse_time("10:10"), time(10, 10))
        self.assertEqual(parse_time("10:20:30.400"), time(10, 20, 30, 400000))
        self.assertEqual(parse_time("10:20:30,400"), time(10, 20, 30, 400000))
        self.assertEqual(parse_time("4:8:16"), time(4, 8, 16))
        # Time zone offset is ignored.
        self.assertEqual(parse_time("00:05:23+04:00"), time(0, 5, 23))
        # Invalid inputs
        self.assertIsNone(parse_time("00:05:"))
        self.assertIsNone(parse_time("00:05:23,"))
        self.assertIsNone(parse_time("00:05:23+"))
        self.assertIsNone(parse_time("00:05:23+25:00"))
        self.assertIsNone(parse_time("4:18:101"))
        self.assertIsNone(parse_time("91500"))
        with self.assertRaises(ValueError):
            parse_time("09:15:90")

    def test_parse_datetime(self):
        valid_inputs = (
            ("2012-04-23T09:15:00", datetime(2012, 4, 23, 9, 15)),
            ("2012-4-9 4:8:16", datetime(2012, 4, 9, 4, 8, 16)),
            (
                "2012-04-23T09:15:00Z",
                datetime(2012, 4, 23, 9, 15, 0, 0, get_fixed_timezone(0)),
            ),
            (
                "2012-4-9 4:8:16-0320",
                datetime(2012, 4, 9, 4, 8, 16, 0, get_fixed_timezone(-200)),
            ),
            (
                "2012-04-23T10:20:30.400+02:30",
                datetime(2012, 4, 23, 10, 20, 30, 400000, get_fixed_timezone(150)),
            ),
            (
                "2012-04-23T10:20:30.400+02",
                datetime(2012, 4, 23, 10, 20, 30, 400000, get_fixed_timezone(120)),
            ),
            (
                "2012-04-23T10:20:30.400-02",
                datetime(2012, 4, 23, 10, 20, 30, 400000, get_fixed_timezone(-120)),
            ),
            (
                "2012-04-23T10:20:30,400-02",
                datetime(2012, 4, 23, 10, 20, 30, 400000, get_fixed_timezone(-120)),
            ),
            (
                "2012-04-23T10:20:30.400 +0230",
                datetime(2012, 4, 23, 10, 20, 30, 400000, get_fixed_timezone(150)),
            ),
            (
                "2012-04-23T10:20:30,400 +00",
                datetime(2012, 4, 23, 10, 20, 30, 400000, get_fixed_timezone(0)),
            ),
            (
                "2012-04-23T10:20:30   -02",
                datetime(2012, 4, 23, 10, 20, 30, 0, get_fixed_timezone(-120)),
            ),
        )
        for source, expected in valid_inputs:
            with self.subTest(source=source):
                self.assertEqual(parse_datetime(source), expected)

        # Invalid inputs
        self.assertIsNone(parse_datetime("20120423091500"))
        with self.assertRaises(ValueError):
            parse_datetime("2012-04-56T09:15:90")


class DurationParseTests(unittest.TestCase):
    def test_parse_python_format(self):
        timedeltas = [
            timedelta(days=4, minutes=15, seconds=30, milliseconds=100),  # fractions of seconds
            timedelta(hours=10, minutes=15, seconds=30),  # hours, minutes, seconds
            timedelta(days=4, minutes=15, seconds=30),  # multiple days
            timedelta(days=1, minutes=00, seconds=00),  # single day
            timedelta(days=-4, minutes=15, seconds=30),  # negative durations
            timedelta(minutes=15, seconds=30),  # minute & seconds
            timedelta(seconds=30),  # seconds
        ]
        for delta in timedeltas:
            with self.subTest(delta=delta):
                self.assertEqual(parse_duration(format(delta)), delta)

    def test_parse_postgresql_format(self):
        test_values = (
            ("1 day", timedelta(1)),
            ("-1 day", timedelta(-1)),
            ("1 day 0:00:01", timedelta(days=1, seconds=1)),
            ("1 day -0:00:01", timedelta(days=1, seconds=-1)),
            ("-1 day -0:00:01", timedelta(days=-1, seconds=-1)),
            ("-1 day +0:00:01", timedelta(days=-1, seconds=1)),
            (
                "4 days 0:15:30.1",
                timedelta(days=4, minutes=15, seconds=30, milliseconds=100),
            ),
            (
                "4 days 0:15:30.0001",
                timedelta(days=4, minutes=15, seconds=30, microseconds=100),
            ),
            ("-4 days -15:00:30", timedelta(days=-4, hours=-15, seconds=-30)),
        )
        for source, expected in test_values:
            with self.subTest(source=source):
                self.assertEqual(parse_duration(source), expected)

    def test_seconds(self):
        self.assertEqual(parse_duration("30"), timedelta(seconds=30))

    def test_minutes_seconds(self):
        self.assertEqual(parse_duration("15:30"), timedelta(minutes=15, seconds=30))
        self.assertEqual(parse_duration("5:30"), timedelta(minutes=5, seconds=30))

    def test_hours_minutes_seconds(self):
        self.assertEqual(parse_duration("10:15:30"), timedelta(hours=10, minutes=15, seconds=30))
        self.assertEqual(parse_duration("1:15:30"), timedelta(hours=1, minutes=15, seconds=30))
        self.assertEqual(
            parse_duration("100:200:300"),
            timedelta(hours=100, minutes=200, seconds=300),
        )

    def test_days(self):
        self.assertEqual(parse_duration("4 15:30"), timedelta(days=4, minutes=15, seconds=30))
        self.assertEqual(
            parse_duration("4 10:15:30"),
            timedelta(days=4, hours=10, minutes=15, seconds=30),
        )

    def test_fractions_of_seconds(self):
        test_values = (
            ("15:30.1", timedelta(minutes=15, seconds=30, milliseconds=100)),
            ("15:30.01", timedelta(minutes=15, seconds=30, milliseconds=10)),
            ("15:30.001", timedelta(minutes=15, seconds=30, milliseconds=1)),
            ("15:30.0001", timedelta(minutes=15, seconds=30, microseconds=100)),
            ("15:30.00001", timedelta(minutes=15, seconds=30, microseconds=10)),
            ("15:30.000001", timedelta(minutes=15, seconds=30, microseconds=1)),
            ("15:30,000001", timedelta(minutes=15, seconds=30, microseconds=1)),
        )
        for source, expected in test_values:
            with self.subTest(source=source):
                self.assertEqual(parse_duration(source), expected)

    def test_negative(self):
        test_values = (
            ("-4 15:30", timedelta(days=-4, minutes=15, seconds=30)),
            ("-172800", timedelta(days=-2)),
            ("-15:30", timedelta(minutes=-15, seconds=-30)),
            ("-1:15:30", timedelta(hours=-1, minutes=-15, seconds=-30)),
            ("-30.1", timedelta(seconds=-30, milliseconds=-100)),
            ("-30,1", timedelta(seconds=-30, milliseconds=-100)),
            ("-00:01:01", timedelta(minutes=-1, seconds=-1)),
            ("-01:01", timedelta(seconds=-61)),
            ("-01:-01", None),
        )
        for source, expected in test_values:
            with self.subTest(source=source):
                self.assertEqual(parse_duration(source), expected)

    def test_iso_8601(self):
        test_values = (
            ("P4Y", None),
            ("P4M", None),
            ("P4W", None),
            ("P4D", timedelta(days=4)),
            ("-P1D", timedelta(days=-1)),
            ("P0.5D", timedelta(hours=12)),
            ("P0,5D", timedelta(hours=12)),
            ("-P0.5D", timedelta(hours=-12)),
            ("-P0,5D", timedelta(hours=-12)),
            ("PT5H", timedelta(hours=5)),
            ("-PT5H", timedelta(hours=-5)),
            ("PT5M", timedelta(minutes=5)),
            ("-PT5M", timedelta(minutes=-5)),
            ("PT5S", timedelta(seconds=5)),
            ("-PT5S", timedelta(seconds=-5)),
            ("PT0.000005S", timedelta(microseconds=5)),
            ("PT0,000005S", timedelta(microseconds=5)),
            ("-PT0.000005S", timedelta(microseconds=-5)),
            ("-PT0,000005S", timedelta(microseconds=-5)),
            ("-P4DT1H", timedelta(days=-4, hours=-1)),
            # Invalid separators for decimal fractions.
            ("P3(3D", None),
            ("PT3)3H", None),
            ("PT3|3M", None),
            ("PT3/3S", None),
        )
        for source, expected in test_values:
            with self.subTest(source=source):
                self.assertEqual(parse_duration(source), expected)



import hishel
import httpx
from patent_client import __version__

session = hishel.AsyncCacheClient(
    headers={
        "Authorization": "OQmPwAN1QD4OXe25jpmMD27zmnM21gIL0lg85G6j",
        "User-Agent": f"Mozilla/5.0 Python Patent Clientbot/{__version__} (parkerhancock@users.noreply.github.com)",
    },
    transport=httpx.AsyncHTTPTransport(
        retries=3,
    ),
    timeout=5 * 60,
)



from __future__ import annotations

from collections import OrderedDict
from copy import deepcopy
from itertools import chain
from typing import AsyncIterator
from typing import Generic
from typing import Iterator
from typing import TypeVar
from typing import Union

from yankee.data import Collection

from ..asyncio_util import run_async_iterator
from ..asyncio_util import run_sync

ModelType = TypeVar("ModelType")


class ManagerConfig:
    """
    Manager Configuration Class

    This class holds configuration information for a manager
    """

    def __init__(self):
        self.filter = OrderedDict()
        self.order_by = list()
        self.options = dict()
        self.limit = None
        self.offset = 0
        self.annotations = list()

    def __eq__(self, other):
        return (
            self.filter == other.filter
            and self.order_by == other.order_by
            and self.options == other.options
            and self.limit == other.limit
            and self.offset == other.offset
            and self.annotations == other.annotations
        )


class Manager(Collection, Generic[ModelType]):
    """
    Manager Class

    This class is essentially a configurable generator. It is intended to be initialized
    as an empty object at Model.objects. Users can then call methods to modify the manager.
    All methods should return a brand-new manager with the appropriate parameters re-set.
    The manager's attributes are stored in a dictionary at Manager.config.

    """

    primary_key: str = ""

    def __init__(self, config=None):
        self.config = config or ManagerConfig()
        if hasattr(self, "__schema__") and callable(self.__schema__):
            self.__schema__ = self.__schema__()

    # Manager Iteration / Slicing

    def __iter__(self) -> Iterator[ModelType]:
        for item in run_async_iterator(self.__aiter__()):
            yield item

    def __aiter__(self) -> AsyncIterator[ModelType]:
        return self._aget_results()

    def __getitem__(self, key: Union[slice, int]) -> Union[Manager[ModelType], ModelType]:
        if isinstance(key, slice):
            if key.step != None:
                raise AttributeError("Step is not supported")
            start = key.start if key.start else 0
            start = len(self) + start if start < 0 else start
            stop = key.stop if key.stop else len(self)
            stop = len(self) + stop if stop < 0 else stop
            mger = self.offset(start + self.config.offset)
            mger = mger.limit(stop - start)
            return mger
        return self.offset(key).first()

    # Basic Manager Attributes

    def __len__(self) -> int:
        # The default len function runs the iterator and counts. There may be
        # more efficient ways to do it for any given subclass, but this is the
        # basic way
        return run_sync(self.alen())

    def __eq__(self, other) -> bool:
        return bool(self.config == other.config and isinstance(self, type(other)))

    def __add__(self, other):
        return Collection(chain(self, other))

    # Manager Modification Functions

    def filter(self, *args, **kwargs) -> Manager[ModelType]:
        """Apply a new filtering condition"""
        mger = deepcopy(self)
        if args:
            kwargs[self.primary_key] = args
        update_values = OrderedDict()
        for key in sorted(kwargs.keys()):
            update_values[key] = kwargs[key]
        mger.config.filter.update(update_values)
        return mger

    def order_by(self, *args) -> Manager[ModelType]:
        """Specify the order that argument should be returned in"""
        mger = deepcopy(self)
        mger.config.order_by = args
        return mger

    def option(self, **kwargs) -> Manager[ModelType]:
        """Set a key:value option on the manager"""
        mger = deepcopy(self)
        mger.config.options = {**mger.config.options, **kwargs}
        return mger

    def limit(self, limit) -> Manager[ModelType]:
        """Limit the number of records that are returned"""
        mger = deepcopy(self)
        mger.config.limit = limit
        return mger

    def offset(self, offset) -> Manager[ModelType]:
        """Specify the number of records from the beginning from which to apply an offset"""
        mger = deepcopy(self)
        mger.config.offset = self.config.offset + offset
        return mger

    async def afirst(self) -> ModelType:
        """Get the first object in the manager"""
        return await anext(self.__aiter__())

    def first(self) -> ModelType:
        """Get the first object in the manager"""
        return run_sync(self.afirst())

    async def aget(self, *args, **kwargs) -> ModelType:
        """If the critera results in a single record, return it, else raise an exception"""
        mger = self.filter(*args, **kwargs)
        length = await mger.alen()
        if length > 1:
            raise ValueError("More than one document found!")
        if length == 0:
            raise ValueError("No documents found!")
        return await mger.afirst()

    def get(self, *args, **kwargs) -> ModelType:
        return run_sync(self.aget(*args, **kwargs))

    # Basic Manager Fetching

    def count(self) -> int:
        """Returns number of records in the QuerySet. Alias for len(self)"""
        return len(self)

    def all(self) -> Manager[ModelType]:
        """Return self. Does nothing"""
        return self



import logging
import re
import warnings
from collections.abc import Sequence
from typing import AsyncIterator

from patent_client.util import Manager
from patent_client.util.request_util import get_start_and_row_count
from urllib3.connectionpool import InsecureRequestWarning

from .api import AssignmentApi
from .model import Assignment

warnings.filterwarnings("ignore", category=InsecureRequestWarning)

NUMBER_CLEAN_RE = re.compile(r"[^\d]")
clean_number = lambda x: NUMBER_CLEAN_RE.sub("", str(x))


logger = logging.getLogger(__name__)


class AssignmentManager(Manager[Assignment]):
    fields = {
        "patent_number": "PatentNumber",
        "appl_id": "ApplicationNumber",
        "app_early_pub_number": "PublicationNumber",
        "assignee": "OwnerName",
        "assignor": "PriorOwnerName",
        "pct_number": "PCTNumber",
        "correspondent": "CorrespondentName",
        "id": "ReelFrame",
    }
    page_size = 100
    obj_class = "patent_client.uspto_assignments.Assignment"
    primary_key = "id"

    def __init__(self, *args, **kwargs):
        super(AssignmentManager, self).__init__(*args, **kwargs)

    @property
    def allowed_filters(self):
        return list(self.fields.keys())

    async def _aget_results(self) -> AsyncIterator[Assignment]:
        for start, rows in get_start_and_row_count(self.config.limit, self.config.offset, self.page_size):
            response = await AssignmentApi.alookup(**{**self.get_query(), "start": start, "rows": rows})
            for doc in response.docs:
                yield doc
            if len(response.docs) < rows:
                break

    def get_query(self):
        """Get assignments.
        Args:
            patent: pat no to search
            application: app no to search
            assignee: assignee name to search
        """

        for key, value in self.config.filter.items():
            field = self.fields[key]
            query = value
        if field in ["PatentNumber", "ApplicationNumber"]:
            if isinstance(query, Sequence) and not isinstance(query, str):
                query = [clean_number(q) for q in query]
            else:
                query = clean_number(query)
        # Handle Ordering
        order_map = {"execution_date": "ExecutionDate+asc", "-execution_date": "ExecutionDate+desc"}
        if len(self.config.order_by) > 1:
            raise ValueError("Assignment API does not support multiple sort parameters!")
        elif len(self.config.order_by) == 1:
            sort = order_map[self.config.order_by[0]]
        else:
            sort = "ExecutionDate+desc"

        if isinstance(query, list):
            query = [f'"{q}"' for q in query]

        query = {
            "filter": field,
            "query": " OR ".join(query) if isinstance(query, Sequence) and not isinstance(query, str) else query,
            "sort": sort,
        }
        return query

    async def alen(self) -> int:
        response = await AssignmentApi.alookup(**self.get_query())
        max_len = response.num_found
        return min(max_len, self.config.limit) if self.config.limit else max_len

    @property
    def query_fields(self):
        fields = self.fields
        for k in sorted(fields.keys()):
            print(k)



import datetime
from dataclasses import dataclass
from dataclasses import field

from patent_client.util.base.model import Model


@dataclass
class DocumentId(Model):
    doc_type: str = None
    id_type: str = None
    country: str = None
    number: str = None
    kind: str = None
    date: "datetime.date" = None
    name: str = None

    def __str__(self):
        return f"{self.country}{self.number}{self.kind}"


@dataclass
class NumberServiceResult(Model):
    input_doc: DocumentId
    output_doc: DocumentId
    service_version: str = None
    messages: list = field(default_factory=list)



import datetime
import re
from collections.abc import Sequence

import lxml.html as ETH
from yankee.json.schema import fields as f
from yankee.json.schema import RegexSchema
from yankee.json.schema import Schema
from yankee.json.schema import ZipSchema

newline_re = re.compile(r"<br />\s*")
bad_break_re = re.compile(r"<br />\s+")


def html_to_text(html):
    html = newline_re.sub("\n\n", html)
    # html = bad_break_re.sub(" ", html)
    return "".join(ETH.fromstring(html).itertext())


class HtmlField(f.Field):
    def deserialize(self, obj):
        obj = super().deserialize(obj)
        if obj is None:
            return None
        elif isinstance(obj, Sequence) and not isinstance(obj, str):
            return "\n\n".join(html_to_text(s) for s in obj)
        else:
            return html_to_text(obj)


def format_appl_id(string):
    if string.startswith("D"):
        string = "29" + string[1:]
    return string.replace("/", "")


class DocumentStructureSchema(Schema):
    number_of_claims = f.Integer("numberOfClaims")
    number_of_drawing_sheets = f.Integer("numberOfDrawingSheets")
    number_of_figures = f.Integer("numberOfFigures")
    page_count = f.Integer("pageCount")
    front_page_end = f.Integer("frontPageEnd")
    front_page_start = f.Integer("frontPageStart")
    bib_start = f.Integer("bibStart")
    bib_end = f.Integer("bibEnd")
    abstract_start = f.Integer("abstractStart")
    abstract_end = f.Integer("abstractEnd")
    drawings_start = f.Integer("drawingsStart")
    drawings_end = f.Integer("drawingsEnd")
    description_start = f.Integer("descriptionStart")
    description_end = f.Integer("descriptionEnd")
    specification_start = f.Integer("specificationStart")
    specification_end = f.Integer("specificationEnd")
    claims_end = f.Integer("claimsEnd")
    claims_start = f.Integer("claimsStart")
    amend_start = f.Integer("amendStart")
    amend_end = f.Integer("amendEnd")
    cert_correction_end = f.Integer("certCorrectionEnd")
    cert_correction_start = f.Integer("certCorrectionStart")
    cert_reexamination_end = f.Integer("certReexaminationEnd")
    cert_reexamination_start = f.Integer("certReexaminationStart")
    ptab_start = f.Integer("ptabStart")
    ptab_end = f.Integer("ptabEnd")
    search_report_start = f.Integer("searchReportStart")
    search_report_end = f.Integer("searchReportEnd")
    supplemental_start = f.Integer("supplementalStart")
    supplemental_end = f.Integer("supplementalEnd")


class PublicSearchSchema(Schema):
    guid = f.String("guid")

    appl_id = f.String("applicationNumber", formatter=format_appl_id)
    app_filing_date = f.Date("applicationFilingDate.0")
    related_appl_filing_date = f.List(f.Date, "relatedApplFilingDate")
    publication_number = f.String("publicationReferenceDocumentNumber")
    kind_code = f.String("kindCode.0")
    publication_date = f.Date("datePublished")
    patent_title = HtmlField("inventionTitle")

    inventors_short = f.String("inventorsShort")
    applicant_name = f.List(f.String, "applicantName")
    assignee_name = f.List(f.String, "assigneeName")
    government_interest = f.List(f.String, "governmentInterest")
    primary_examiner = f.String("primaryExaminer")
    assistant_examiner = f.List(f.String, "assistantExaminer")

    main_classification_code = f.String("mainClassificationCode")
    cpc_additional = f.DelimitedString(f.Str(), "cpcAdditionalFlattened", delimeter=";")
    cpc_inventive = f.DelimitedString(f.Str(), "cpcInventiveFlattened", delimeter=";")
    ipc_code = f.DelimitedString(f.Str(), "ipcCodeFlattened", delimeter=";")
    uspc_full_classification = f.DelimitedString(f.Str(), "uspcFullClassificationFlattened", delimeter=";")

    image_file_name = f.String("imageFileName")
    image_location = f.String("imageLocation")
    document_structure = DocumentStructureSchema(data_key=False)

    type = f.String("type")
    database_name = f.String("databaseName")
    composite_id = f.String("compositeId")
    document_id = f.String("documentId")
    document_size = f.Integer("documentSize")
    family_identifier_cur = f.Integer("familyIdentifierCur")
    language_indicator = f.String("languageIndicator")

    score = f.Float("score")


class SpecificationField(f.Combine):
    pass


class DocumentSchema(Schema):
    abstract = HtmlField("abstractHtml")
    government_interest = HtmlField("governmentInterest")
    background = HtmlField("backgroundTextHtml")
    brief = HtmlField("briefHtml")
    description = HtmlField("descriptionHtml")
    claim_statement = f.String("claimStatement")
    claims = HtmlField("claimsHtml")


class UsReferenceSchema(ZipSchema):
    publication_number = f.String("urpn")
    # us_class = f.String("usRefClassification")
    # cpc_class = f.String("usRefCpcClassification")
    # group = f.String("usRefGroup")
    pub_month = f.Date(
        "usRefIssueDate",
        dt_converter=lambda s: datetime.datetime(year=int(s[:4]), month=int(s[4:6]), day=1),
    )
    patentee_name = f.String("usRefPatenteeName")
    cited_by_examiner = f.Boolean("usRefGroup", true_func=lambda s: "examiner" in s)


class ForeignReferenceSchema(ZipSchema):
    citation_classification = f.String("foreignRefCitationClassification")
    citation_cpc = f.String("foreignRefCitationCpc")
    country_code = f.String("foreignRefCountryCode")
    # group = f.String("foreignRefGroup")
    patent_number = f.String("foreignRefPatentNumber")
    pub_month = f.Date(
        "foreignRefPubDate",
        dt_converter=lambda s: datetime.datetime(year=int(s[:4]), month=int(s[4:6]), day=1),
    )
    cited_by_examiner = f.Boolean("foreignRefGroup", true_func=lambda s: "examiner" in s)


class NplReferenceSchema(RegexSchema):
    __regex__ = r"(?P<citation>.*)(?P<cited_by_examiner>cited by (applicant|examiner).?$)"
    citation = f.String()
    cited_by_examiner = f.Bool(true_func=lambda s: "examiner" in s)


class RelatedApplicationSchema(ZipSchema):
    child_patent_country = f.String("relatedApplChildPatentCountry")
    child_patent_number = f.String("relatedApplChildPatentNumber")
    country_code = f.String("relatedApplCountryCode")
    filing_date = f.Date("relatedApplFilingDate")
    number = f.String("relatedApplNumber")
    parent_status_code = f.String("relatedApplParentStatusCode")
    patent_issue_date = f.Date("relatedApplPatentIssueDate")
    patent_number = f.String("relatedApplPatentNumber")


class InventorSchema(ZipSchema):
    name = f.String("inventorsName")
    city = f.String("inventorCity")
    country = f.String("inventorCountry")
    postal_code = f.String("inventorPostalCode")
    state = f.String("inventorState")


class ApplicantSchema(ZipSchema):
    city = f.String("applicantCity")
    country = f.String("applicantCountry")
    # group = f.String("applicantGroup")
    name = f.String("applicantName")
    state = f.String("applicantState")
    zip_code = f.String("applicantZipCode")
    authority_type = f.String("applicantAuthorityType")


class AssigneeSchema(ZipSchema):
    city = f.String("assigneeCity")
    country = f.String("assigneeCountry")
    name = f.String("assigneeName")
    postal_code = f.String("assigneePostalCode")
    state = f.String("assigneeState")
    type_code = f.String("assigneeTypeCode")


class CpcCodeSchema(RegexSchema):
    __regex__ = r"(?P<cpc_class>.{4})(?P<cpc_subclass>[^ ]+) (?P<version>\d{8})"
    cpc_class = f.Str()
    cpc_subclass = f.Str()
    version = f.Date()


class IntlCodeSchema(RegexSchema):
    __regex__ = r"(?P<intl_class>.{4})(?P<intl_subclass>[^ ]+) (?P<version>\d{8})"
    intl_class = f.Str()
    intl_subclass = f.Str()
    version = f.Date()


class ForeignPriorityApplicationSchema(ZipSchema):
    country = f.Str("priorityClaimsCountry")
    app_filing_date = f.Date("priorityClaimsDate")
    app_number = f.Str("priorityClaimsDocNumber")


class PublicSearchDocumentSchema(Schema):
    guid = f.String("guid")
    publication_number = f.String("pubRefDocNumber")
    publication_date = f.Date("datePublished")

    appl_id = f.String("applicationNumber", formatter=format_appl_id)
    patent_title = HtmlField("inventionTitle")
    app_filing_date = f.Date("applicationFilingDate.0")
    application_type = f.String("applicationRefFilingType")
    family_identifier_cur = f.Integer("familyIdentifierCur")
    related_apps = RelatedApplicationSchema(data_key=False)
    foreign_priority = ForeignPriorityApplicationSchema(data_key=False)
    type = f.String("type")

    # Parties
    inventors = InventorSchema(data_key=False)
    inventors_short = f.String("inventorsShort")
    applicants = ApplicantSchema(data_key=False)
    assignees = AssigneeSchema(data_key=False)

    group_art_unit = f.String("examinerGroup")
    primary_examiner = f.String("primaryExaminer")
    assistant_examiner = f.List(f.String, "assistantExaminer")
    legal_firm_name = f.List(f.String, "legalFirmName")
    attorney_name = f.List(f.String, "attorneyName")

    # Text Data
    document = DocumentSchema(data_key=False)
    document_structure = DocumentStructureSchema(data_key=False)

    # Image Data
    image_file_name = f.String("imageFileName")
    image_location = f.String("imageLocation")

    # Metadata
    composite_id = f.String("compositeId")
    database_name = f.String("databaseName")
    derwent_week_int = f.Integer("derwentWeekInt")

    # References Cited
    us_references = UsReferenceSchema(data_key=False)
    foreign_references = ForeignReferenceSchema(data_key=False)
    npl_references = f.DelimitedString(NplReferenceSchema, "otherRefPub.0", delimeter="<br />")

    # Classifications
    cpc_inventive = f.List(CpcCodeSchema)
    cpc_additional = f.List(CpcCodeSchema)

    intl_class_issued = f.DelimitedString(f.String, "ipcCodeFlattened", delimeter=";")
    intl_class_current_primary = f.List(IntlCodeSchema, "curIntlPatentClassificationPrimary")
    intl_class_currrent_secondary = f.List(IntlCodeSchema, "curIntlPatentClassificationSecondary")

    us_class_current = f.DelimitedString(f.Str(), "uspcFullClassificationFlattened", delimeter=";")
    us_class_issued = f.List(f.Str, "issuedUsClassificationFull")

    field_of_search_us = f.List(f.Str(), "fieldOfSearchClassSubclassHighlights")
    field_of_search_cpc = f.List(f.Str(), "fieldOfSearchCpcClassification")



import re

whitespace_except_newlines_re = re.compile(r"[ \t\r\f\v]+")
newlines_re = re.compile(r"\n+")


def clean_whitespace(string, preserve_newlines=False):
    string = string.strip()
    string = whitespace_except_newlines_re.sub(" ", string)
    if preserve_newlines:
        string = newlines_re.sub("\n", string)
    else:
        string = newlines_re.sub(" ", string)
    return string


non_digit_re = re.compile(r"[^\d]+")


def clean_number(string):
    return non_digit_re.sub("", string)


def clean_appl_id(string):
    return string.replace(",", "").replace("/", "").replace("D", "29").strip()



from .biblio import BiblioResultSchema
from .fulltext import ClaimsSchema
from .fulltext import DescriptionSchema
from .images import ImagesSchema
from .search import SearchSchema

__all__ = [
    "BiblioResultSchema",
    "ClaimsSchema",
    "DescriptionSchema",
    "ImagesSchema",
    "SearchSchema",
]



import logging
from io import BytesIO

from patent_client.epo.ops.session import asession
from yankee.data import AttrDict


logger = logging.getLogger(__name__)


class PublishedBiblioAsyncApi:
    @classmethod
    async def get_constituents(cls, number, doc_type="publication", format="docdb", constituents=("biblio",)):
        """Published Data Constituents API
        number: document number to search
        doc_type: document type (application / publication)
        format: document number format (original / docdb / epodoc)
        constituents: what data to retrieve. Can be combined. (biblio / abstract / full-cycle)

        """
        base_url = f"http://ops.epo.org/3.2/rest-services/published-data/{doc_type}/{format}/{number}/"
        if isinstance(constituents, str):
            constituents = (constituents,)
        url = base_url + ",".join(constituents)
        response = await asession.get(url)
        response.raise_for_status()
        return response.text

    @classmethod
    async def get_biblio(cls, number, doc_type="publication", format="docdb"):
        return await cls.get_constituents(number, doc_type, format, constituents="biblio")

    @classmethod
    async def get_abstract(cls, number, doc_type="publication", format="docdb"):
        return await cls.get_constituents(number, doc_type, format, constituents="abstract")

    @classmethod
    async def get_full_cycle(cls, number, doc_type="publication", format="docdb"):
        return await cls.get_constituents(number, doc_type, format, constituents="full-cycle")


class PublishedFulltextAsyncApi:
    fulltext_jurisdictions = "EP, WO, AT, BE, BG, CA, CH, CY, CZ, DK, EE, ES, FR, GB, GR, HR, IE, IT, LT, LU, MC, MD, ME, NO, PL, PT, RO, RS, SE, SK".split(
        ", "
    )

    @classmethod
    async def get_fulltext_result(cls, number, doc_type="publication", format="docdb", inquiry="fulltext"):
        """Published Fulltext API
        number: document number to search
        doc_type: document type (application / publication)
        format: document number format (original / docdb / epodoc)
        inquiry: what data to retrieve. Can be combined. (fulltext / description / claims)

        """
        url = f"http://ops.epo.org/3.2/rest-services/published-data/{doc_type}/{format}/{number}/{inquiry}"
        if number[:2] not in cls.fulltext_jurisdictions:
            raise ValueError(
                f"Fulltext Is Not Available For Country Code {number[:2]}. Fulltext is only available in {', '.join(cls.fulltext_jurisdictions)}"
            )
        response = await asession.get(url)
        response.raise_for_status
        return response.text

    @classmethod
    async def get_description(cls, number, doc_type="publication", format="docdb"):
        return await cls.get_fulltext_result(number, doc_type="publication", format="docdb", inquiry="description")

    @classmethod
    async def get_claims(cls, number, doc_type="publication", format="docdb"):
        return await cls.get_fulltext_result(number, doc_type="publication", format="docdb", inquiry="claims")


class PublishedSearchAsyncApi:
    @classmethod
    async def search(cls, query, start=1, end=100):
        base_url = "http://ops.epo.org/3.2/rest-services/published-data/search"
        range = f"{start}-{end}"
        logger.debug(f"OPS Search Endpoint - Query: {query}\nRange: {start}-{end}")
        response = await asession.get(base_url, params={"Range": range, "q": query})
        if response.status_code == 404:
            return AttrDict.convert(
                {
                    "query": "query",
                    "num_results": 0,
                    "begin": start,
                    "end": end,
                    "results": list(),
                }
            )
        response.raise_for_status()
        return response.text


class PublishedImagesAsyncApi:
    @classmethod
    async def get_images(cls, number, doc_type="publication", format="docdb"):
        base_url = f"http://ops.epo.org/3.2/rest-services/published-data/{doc_type}/{format}/{number}/images"
        response = await asession.get(base_url)
        response.raise_for_status()
        return response.text

    @classmethod
    async def get_page_image(cls, country, number, kind, image_type, page_number, image_format="pdf"):
        response = await asession.get(
            f"https://ops.epo.org/3.2/rest-services/published-data/images/{country}/{number}/{kind}/{image_type}.{image_format}",
            params={"Range": page_number},
            stream=True,
        )
        response.raise_for_status()
        return BytesIO(response.raw.read())

    @classmethod
    async def get_page_image_from_link(cls, link, page_number, image_format="pdf"):
        response = await asession.get(
            f"https://ops.epo.org/3.2/rest-services/{link}.{image_format}",
            params={"Range": page_number},
            stream=True,
        )
        response.raise_for_status()
        return BytesIO(response.raw.read())


class PublishedAsyncApi:
    biblio = PublishedBiblioAsyncApi
    fulltext = PublishedFulltextAsyncApi
    search = PublishedSearchAsyncApi
    images = PublishedImagesAsyncApi



from yankee.base.schema import ListCollection

from .model import Assignment
from .schema import AssignmentPageSchema
from .session import session


allowed_filters = [
    "PCTNumber",
    "OwnerName",
    "CorrespondentName",
    "PriorOwnerName",
    "ApplicationNumber",
    "PatentNumber",
    "PublicationNumber",
    "IntlRegistrationNumber",
    "ReelFrame",
]

allowed_sorts = ["ExecutionDate+desc", "ExecutionDate+asc"]


def validate_input(input, input_name, allowed_values):
    if input not in allowed_values:
        raise ValueError(f"{input_name} must be one of {allowed_values}")


class AssignmentApi:
    @classmethod
    async def alookup(
        self, query: str, filter: str, rows=8, start=0, sort="ExecutionDate+desc"
    ) -> ListCollection[Assignment]:
        # Because of their limited utility, we omit fields, highlight, and facet
        url = "https://assignment-api.uspto.gov/patent/lookup"
        validate_input(filter, "filter", allowed_filters)
        validate_input(sort, "sort", allowed_sorts)
        response = await session.get(
            url,
            params={
                "filter": filter,
                "query": query,
                "rows": rows,
                "start": start,
                "sort": sort,
                "facet": "false",
            },
            headers={"Accept": "application/xml"},
        )
        response.raise_for_status()
        return AssignmentPageSchema().load(response.text)



import datetime
from collections import defaultdict
from typing import *

from dateutil.parser import ParserError
from patent_client.util.format import clean_whitespace
from yankee.json.schema import fields as f
from yankee.json.schema import Schema


class UsptoDate(f.Date):
    def deserialize(self, elem) -> "Optional[datetime.date]":
        try:
            return super().deserialize(elem)
        except ParserError:
            return datetime.datetime(1900, 1, 1).date()


class InventorNameField(f.Combine):
    name_line_one = f.Str()
    name_line_two = f.Str()
    suffix = f.Str()

    def combine_func(self, obj):
        return clean_whitespace(
            f"{obj.get('name_line_one', '')}; {obj.get('name_line_two', '')} {obj.get('suffix', '')}"
        )


class InventorAddressField(f.Combine):
    street_one = f.Str()
    street_two = f.Str()
    city = f.Str()
    geo_code = f.Str()
    postal_code = f.Str()
    country = f.Str()

    def combine_func(self, obj):
        return clean_whitespace(
            f'{obj.street_one or ""}\n{obj.street_two or ""}\n{obj.city or ""}, {obj.geo_code or ""} {obj.postal_code or ""} {obj.country or ""}',
            preserve_newlines=True,
        )


class PersonSchema(Schema):
    name = InventorNameField(data_key=False)
    address = InventorAddressField(data_key=False)
    rank_no = f.Int()


class InventorSchema(PersonSchema):
    pass


class ApplicantSchema(PersonSchema):
    cust_no = f.Str()


class TransactionSchema(Schema):
    date = UsptoDate("recordDate")
    code = f.Str()
    description = f.Str()


class ChildSchema(Schema):
    __model_name__ = "Relationship"
    parent_appl_id = f.Str("applicationNumberText")
    child_app_filing_date = UsptoDate("filingDate")
    child_app_status = f.Str("applicationStatus")
    child_appl_id = f.Str("claimApplicationNumberText")
    relationship = f.Str(
        "applicationStatusDescription",
        formatter=lambda x: x.replace("This application ", ""),
    )


class ParentSchema(Schema):
    __model_name__ = "Relationship"
    parent_appl_id = f.Str("claimApplicationNumberText")
    child_appl_id = f.Str("applicationNumberText")
    parent_app_filing_date = UsptoDate("filingDate")
    parent_app_status = f.Str("applicationStatus")
    relationship = f.Str(
        "applicationStatusDescription",
        formatter=lambda x: x.replace("This application ", ""),
    )


class OptionalFloat(f.Int):
    def deserialize(self, elem) -> "Optional[int]":
        try:
            return super().deserialize(elem)
        except ValueError:
            return None


class PtaPteHistorySchema(Schema):
    date = UsptoDate("ptaOrPteDate")
    description = f.String("contentsDescription")
    number = f.Float()
    pto_days = OptionalFloat()
    applicant_days = OptionalFloat()
    start = f.Float()


class PtaPteSummarySchema(Schema):
    a_delay = f.Int()
    b_delay = f.Int()
    c_delay = f.Int()
    overlap_delay = f.Int()
    pto_delay = f.Int()
    applicant_delay = f.Int("applDelay")
    pto_adjustments = f.Int()
    total_days = f.Int("totalPtoDays")
    kind = f.Str("ptaPteInd")


class CorrespondentNameSchema(f.Combine):
    line_one = f.Str("corrAddrNameLineOne")
    line_two = f.Str("corrAddrNameLineTwo")

    def combine_func(self, obj):
        obj = defaultdict(lambda: "", obj.to_dict())
        return "{line_one}\n{line_two}".format_map(obj)


class CorrespondentAddressSchema(f.Combine):
    street_1 = f.Str("corrAddrStreetLineOne")
    street_2 = f.Str("corrAddrStreetLineTwo")
    city = f.Str("corrAddrCity")
    geo_code = f.Str("corrAddrGeoRegionCode")
    postal_code = f.Str("corrAddrPostalCode")
    country = f.Str(data_key="corrAddrCountryCd")

    def combine_func(self, obj):
        obj = defaultdict(lambda: "", obj.to_dict())
        return clean_whitespace(
            "{street_1}\n{street_2}\n{city}, {geo_code} {postal_code} {country}".format_map(obj),
            preserve_newlines=True,
        )


class CorrespondentSchema(Schema):
    name = CorrespondentNameSchema(data_key=False)
    address = CorrespondentAddressSchema(data_key=False)
    cust_no = f.Str("corrAddrCustNo")


class AttorneySchema(Schema):
    registration_no = f.Str()
    name = f.Str("fullName")
    phone_num = f.Str()
    reg_status = f.Str()


class ForeignPrioritySchema(Schema):
    priority_claim = f.Str()
    country_name = f.Str()
    filing_date = UsptoDate(dt_format="%m-%d-%Y")


class DocumentSchema(Schema):
    access_level_category = f.Str()
    appl_id = f.Str("applicationNumberText")
    category = f.Str("documentCategory")
    code = f.Str(data_key="documentCode")
    description = f.Str("documentDescription")
    identifier = f.Str("documentIdentifier")
    mail_room_date = UsptoDate()
    page_count = f.Int()
    url = f.Str(data_key="pdf_url")


class ReelFrameField(f.Combine):
    reel_number = f.Str()
    frame_number = f.Str()

    def combine_func(self, obj):
        return f"{obj.reel_number}/{obj.frame_number}"


class AddressField(f.Combine):
    line_1 = f.Str("addressLineOneText", null_value="null")
    line_2 = f.Str("addressLineTwoText", null_value="null")
    line_3 = f.Str("addressLineThreeText", null_value="null")
    line_4 = f.Str("addressLineFourText", null_value="null")

    def combine_func(self, obj):
        return clean_whitespace(
            f"{obj.get('line_1', '')}\n{obj.get('line_2', '')}\n{obj.get('line_3', '')}\n{obj.get('line_4', '')}".strip(),
            preserve_newlines=True,
        )


class AssignorSchema(Schema):
    name = f.Str("assignorName")
    exec_date = UsptoDate()


class AssigneeAddressField(f.Combine):
    line_1 = f.Str("streetLineOneText", null_value="null")
    line_2 = f.Str("streetLineTwoText", null_value="null")
    city = f.Str("cityName")
    country = f.Str("countryCode")
    postal_code = f.Str("postalCode")

    def combine_func(self, obj):
        return clean_whitespace(
            f"{obj.get('line_1', '')}\n{obj.get('line_2', '')}\n{obj.get('city', '')}, {obj.get('country', '')} {obj.get('postal_code', '')}",
            preserve_newlines=True,
        )


class AssigneeSchema(Schema):
    name = f.Str("assigneeName")
    address = AssigneeAddressField(data_key=False)


class AssignmentSchema(Schema):
    id = ReelFrameField(data_key=False)
    correspondent = f.Str("addressNameText")
    correspondent_address = AddressField(data_key=False)
    mail_date = UsptoDate(null_value="NONE")
    received_date = UsptoDate(null_value="NONE")
    recorded_date = UsptoDate(null_value="NONE")
    pages = f.Int("pagesCount")
    conveyance_text = f.Str(
        "converyanceName",
        formatter=lambda x: x.replace(" (SEE DOCUMENT FOR DETAILS).", ""),
    )
    sequence_number = f.Int()
    assignors = f.List(AssignorSchema, "assignors")
    assignees = f.List(AssigneeSchema, "assignee")


class USApplicationSchema(Schema):
    # Basic Bibliographic Data
    appl_id = f.Str()
    app_confr_number = f.Str()
    app_filing_date = UsptoDate()
    app_location = f.Str()
    app_type = f.Str()
    app_entity_status = f.Str()
    app_cls_sub_cls = f.Str()
    app_grp_art_number = f.Str()
    app_exam_name = f.Str()
    first_inventor_file = f.Bool(true_value="Yes")
    patent_title = f.Str()
    # Status Information
    app_status = f.Str()
    app_status_date = UsptoDate()
    # Publication Information
    app_early_pub_number = f.Str()
    app_early_pub_date = UsptoDate()
    patent_number = f.Str()
    patent_issue_date = UsptoDate()
    wipo_early_pub_number = f.Str()
    wipo_early_pub_date = UsptoDate()
    # Correspondent / Attorney Information
    corr_addr_cust_no = f.Str()
    app_cust_number = f.Str()
    app_attr_dock_number = f.Str()

    # Assignments
    assignments = f.List(AssignmentSchema)

    # Parties
    inventors = f.List(InventorSchema)
    applicants = f.List(ApplicantSchema)
    transactions = f.List(TransactionSchema)
    child_continuity = f.List(ChildSchema)
    parent_continuity = f.List(ParentSchema)
    pta_pte_tran_history = f.List(PtaPteHistorySchema)
    pta_pte_summary = PtaPteSummarySchema(data_key=False)
    correspondent = CorrespondentSchema(data_key=False)
    attorneys = f.List(AttorneySchema, data_key="attrnyAddr")
    foreign_priority = f.List(ForeignPrioritySchema)


class PedsPageSchema(Schema):
    index_last_updated = UsptoDate("queryResults.indexLastUpdatedDate")
    num_found = f.Int("queryResults.searchResponse.response.numFound")
    applications = f.List(USApplicationSchema, "queryResults.searchResponse.response.docs")



from typing import Generic
from typing import TypeVar

from patent_client.util.base.manager import Manager

from . import public_search_async_api
from .model import Patent
from .model import PatentBiblio
from .model import PublicSearch
from .model import PublicSearchDocument
from .model import PublishedApplication
from .model import PublishedApplicationBiblio
from .query import QueryBuilder
from .schema import PublicSearchDocumentSchema
from .schema import PublicSearchSchema


class CapacityException(Exception):
    pass


class FinishedException(Exception):
    pass


T = TypeVar("T")


class GenericPublicSearchManager(Manager, Generic[T]):
    __schema__ = PublicSearchSchema
    page_size = 500
    primary_key = "patent_number"
    query_builder = QueryBuilder()

    async def _aget_results(self):
        query = self._query
        order_by = self._order_by
        sources = self.config.options.get("sources", ["US-PGPUB", "USPAT", "USOCR"])
        page_no = 0
        obj_counter = 0
        try:
            while True:
                page = await public_search_async_api.run_query(
                    query=query,
                    start=page_no * self.page_size,
                    limit=self.page_size,
                    sort=order_by,
                    sources=sources,
                )
                for obj in page["patents"]:
                    if self.config.limit and obj_counter >= self.config.limit + self.config.offset:
                        raise FinishedException()
                    if obj_counter >= self.config.offset:
                        yield self.__schema__.load(obj)
                    obj_counter += 1
                page_no += 1
                if len(page["patents"]) < self.page_size:
                    raise FinishedException()
        except FinishedException:
            pass

    @property
    def _query(self):
        return self.query_builder.build_query(self.config)

    @property
    def _order_by(self):
        return self.query_builder.build_order_by(self.config)

    @property
    def query_fields(self):
        return self.query_builder.search_keywords

    @property
    def order_by_fields(self):
        return self.query_builder.order_by_keywords

    async def alen(self):
        if hasattr(self, "_len"):
            return self._len
        query = self._query
        order_by = self._order_by
        sources = self.config.options.get("sources", ["US-PGPUB", "USPAT", "USOCR"])
        page = await public_search_async_api.run_query(
            query=query, start=0, limit=self.page_size, sort=order_by, sources=sources
        )

        total_results = page["totalResults"]
        total_results -= self.config.offset
        if self.config.limit:
            self._len = min(self.config.limit, total_results)
        else:
            self._len = total_results
        return self._len


class GenericPublicSearchDocumentManager(GenericPublicSearchManager, Generic[T]):
    __doc_schema__ = PublicSearchDocumentSchema()

    async def _aget_results(self):
        result_count = super().__len__()
        if result_count > 20:
            raise CapacityException(
                f"Query would result in more than 20 results! ({result_count} > 20).\nPlease use the associated Biblio method to reduce load on the API (PublicSearch / PatentBiblio / PublishedApplicationBiblio"
            )

        async for obj in super()._aget_results():
            doc = await public_search_async_api.get_document(obj)
            yield self.__doc_schema__.load(doc)


class PublicSearchManager(GenericPublicSearchManager[PublicSearch]):
    pass


class PublicSearchDocumentManager(GenericPublicSearchDocumentManager[PublicSearchDocument]):
    pass


class PatentBiblioManager(GenericPublicSearchManager[PatentBiblio]):
    def __init__(self, config=None):
        super().__init__(config=config)
        self.config.options["sources"] = [
            "USPAT",
        ]


class PatentManager(GenericPublicSearchDocumentManager[Patent]):
    def __init__(self, config=None):
        super().__init__(config=config)
        self.config.options["sources"] = [
            "USPAT",
        ]


class PublishedApplicationBiblioManager(GenericPublicSearchManager[PublishedApplicationBiblio]):
    def __init__(self, config=None):
        super().__init__(config=config)
        self.config.options["sources"] = [
            "US-PGPUB",
        ]


class PublishedApplicationManager(GenericPublicSearchDocumentManager[PublishedApplication]):
    def __init__(self, config=None):
        super().__init__(config=config)
        self.config.options["sources"] = [
            "US-PGPUB",
        ]



import datetime
import logging
import re
import sqlite3
from pathlib import Path

import lxml.etree as ET
from openpyxl import load_workbook
from patent_client import SETTINGS
from patent_client.epo.ops.session import asession

dir = Path(SETTINGS.DEFAULT.BASE_DIR).expanduser() / "epo"
dir.mkdir(exist_ok=True, parents=True)
db_location = dir / "legal_codes.sqlite"


logger = logging.getLogger(__name__)


def current_date():
    return datetime.datetime.now().date()


def generate_legal_code_db():
    current = has_current_spreadsheet()
    if current:
        logger.debug("Legal Code Database is Current - skipping database creation")
        return
    else:
        logger.debug("Legal Code Database is out of date - creating legal code database")
        path = get_spreadsheet()
        create_code_database(path)


def has_current_spreadsheet():
    con = sqlite3.connect(db_location, timeout=30)
    cur = con.cursor()
    try:
        fname = cur.execute("SELECT * FROM meta").fetchone()[0]
        date_string = re.search(r"legal_code_descriptions_(\d+)\.xlsx", fname).group(1)
        try:
            date = datetime.datetime.strptime(date_string, "%Y%W").date()
        except ValueError:
            date = datetime.datetime.strptime(date_string, "%Y%m%d").date()
        age = datetime.datetime.now().date() - date
        logger.debug(f"Legal Code Database is {age} days old")
        return age.days <= 30
    except (sqlite3.OperationalError, TypeError):
        return False


def get_spreadsheet():
    url = "https://www.epo.org/searching-for-patents/data/coverage/weekly.html"
    response = asession.sync.get(url)
    response.raise_for_status()
    tree = ET.HTML(response.text)
    try:
        excel_url = tree.xpath('.//*[contains(@href, "legal_code_descriptions")][1]/@href')[0]
        out_path = dir / excel_url.split("/")[-1]
        if out_path.exists():
            return out_path
        response = asession.sync.get(excel_url, stream=True)
        with out_path.open("wb") as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)
        return out_path
    except Exception:
        logger.debug("Could not find live code file - falling back to default dated 2022-11-12")
        return Path(__file__).parent / "legal_code_descriptions_20221112.xlsx"


def create_code_database(excel_path):
    con = sqlite3.connect(db_location, timeout=30)
    cur = con.cursor()
    try:
        meta = cur.execute("SELECT * FROM meta").fetchone()[0]
        if meta == excel_path.name:
            logger.debug(f"Excel file {excel_path.name} already loaded. Skipping!")
            return
    except (sqlite3.OperationalError, TypeError):
        pass

    cur.execute("CREATE TABLE IF NOT EXISTS meta (file_name text)")
    cur.execute("INSERT INTO meta values (?)", (excel_path.name,))
    wb = load_workbook(excel_path)
    data = list(tuple(i.strip() for i in r) for r in wb[wb.sheetnames[0]].iter_rows(values_only=True))
    rows = data[1:]
    cur.execute(
        """CREATE TABLE IF NOT EXISTS legal_codes (
    country_code text,
    event_code text,
    date_created text,
    influence text,
    description text,
    last_update text,
    description_orig text,
    last_update_orig text,
    event_class text,
    event_class_description text)"""
    )
    cur.execute("""CREATE INDEX IF NOT EXISTS country_event_code ON legal_codes (country_code, event_code)""")
    cur.executemany("INSERT INTO legal_codes values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)", rows)
    con.commit()


class LegalCodes:
    def __init__(self):
        self.connection = sqlite3.connect(db_location, timeout=30)
        self.connection.row_factory = sqlite3.Row

    def get_code_data(self, country_code, legal_code):
        cur = self.connection.cursor()
        try:
            return dict(
                cur.execute(
                    "SELECT * FROM legal_codes WHERE country_code = ? AND event_code = ?",
                    (country_code, legal_code),
                ).fetchone()
            )
        except TypeError:
            raise Exception(f"No Event Data found for {country_code} - {legal_code}")



import lxml.etree as ET
from patent_client.epo.ops.session import asession

from .model import NumberServiceResult
from .schema import NumberServiceResultSchema


class NumberServiceException(Exception):
    pass


async def aconvert_number(
    number, doc_type="publication", input_format="original", output_format="docdb"
) -> NumberServiceResult:
    """Number Conversion Service
    number: number to convert. Must include at least country and doc number,
        can optionally include kind code or publication date. e.g.
        MD.20050130
        MD.20050130.A
        MD.20050130.A.20050130
        JP.2006-147056.A.20060526
        JP.2006147056.A.20060526
        US.(08/921,321).A.19970829
    doc_type: type of document (publication / application / priority)
    input_format: input type (original / docdb / epodoc)
    output_format: output type (original / docdb / epodoc)

    """
    response = await asession.get(
        f"http://ops.epo.org/3.2/rest-services/number-service/{doc_type}/{input_format}/{number}/{output_format}"
    )
    tree = ET.fromstring(response.text.encode())
    result = NumberServiceResultSchema().load(tree)
    errors = [m for m in result.messages if m.kind == "ERROR"]
    if errors:
        raise NumberServiceException("\n".join(str(e) for e in errors))
    return result



from dataclasses import dataclass
from dataclasses import field
from typing import List

from patent_client.util import Model

from ...util import InpadocModel


@dataclass
class Inpadoc(InpadocModel):
    __manager__ = "patent_client.epo.ops.published.manager.SearchManager"
    family_id: str = None
    id_type: str = None
    country: str = None
    doc_number: str = None
    kind: str = None

    @property
    def docdb_number(self):
        return f"{self.country}{self.doc_number}{self.kind}"

    def download(self, path="."):
        return self.images.full_document.download(path)


@dataclass
class Search(Model):
    query: str = None
    num_results: int = None
    begin: int = None
    end: int = None
    results: List[Inpadoc] = field(default_factory=list)



from patent_client.epo.ops.util import Schema
from yankee.xml import fields as f

from . import error_dir


class TagField(f.Field):
    def deserialize(self, obj):
        obj = super().deserialize(obj)
        if obj is None:
            return None
        return obj.tag.split("}")[1].split("-")[0]


class DocumentIdSchema(Schema):
    doc_type = TagField("./*")
    id_type = f.Str(".//@document-id-type")
    country = f.Str(".//epo:country")
    number = f.Str(".//epo:doc-number")
    kind = f.Str(".//epo:kind")
    date = f.Date(".//epo:date")
    name = f.Str(".//epo:name")


def get_messages(text):
    return [error_dir[k] for k in text.split() if k != "SUCCESS"]


class NumberServiceResultSchema(Schema):
    input_doc = DocumentIdSchema(".//ops:input")
    output_doc = DocumentIdSchema(".//ops:output")
    service_version = f.Str('.//ops:meta[@name="version"]/@value')
    messages = f.Str('.//ops:meta[@name="status"]/@value', formatter=get_messages)



import datetime
from typing import *

from yankee.xml.schema import fields as f
from yankee.xml.schema import Schema
from yankee.xml.schema import ZipSchema


# Schemas
class Str(f.String):
    def deserialize(self, elem) -> "Optional[str]":
        result = super().deserialize(elem)
        return result if result != "NULL" else None


class Date(f.Date):
    def deserialize(self, elem) -> "Optional[datetime.date]":
        result = super().deserialize(elem)
        return result if result != datetime.date(1, 1, 1) else None


class PropertySchema(ZipSchema):
    invention_title = Str(".//inventionTitle/str")
    inventors = Str(".//inventors/str")
    # Numbers
    appl_id = Str(".//applNum/str")
    pct_num = Str(".//pctNum/str")
    intl_reg_num = Str(".//intlRegNum/str")
    publ_num = Str(".//publNum/str")
    pat_num = Str(".//patNum/str")
    # Dates
    filing_date = Date(".//filingDate/date")
    intl_publ_date = Date(".//intlPublDate/date")
    issue_date = Date(".//issueDate/date")
    publ_date = Date(".//publDate/date")


class AssignorSchema(ZipSchema):
    name = Str(".//patAssignorName/str")
    ex_date = Date(".//patAssignorExDate/date")
    date_ack = Date(".//patAssignorDateAck/date")


class AssigneeSchema(ZipSchema):
    name = Str(".//patAssigneeName/str")
    line_1 = Str(".//patAssigneeAddress1/str")
    line_2 = Str(".//patAssigneeAddress2/str")
    city = Str(".//patAssigneeCity/str")
    state = Str(".//patAssigneeState/str")
    post_code = Str(".//patAssigneePostcode/str")
    country = Str(".//patAssigneeCountryName/str")

    def post_load(self, obj):
        return [{"name": o.name, "address": self.combine_func(o)} for o in obj]

    def combine_func(self, obj):
        address = f"{obj.get('line_1', '')}\n{obj.get('line_2', '')}".strip()
        if "city" in obj:
            address += f"\n{obj.get('city', '')}, {obj.get('state', '')} {obj.get('post_code', '')}".rstrip()
        if "country" in obj:
            address += f" ({obj.get('country', '')})"
        return address


class CorrespondentAddressField(f.Combine):
    corrAddress1 = Str(".//corrAddress1")
    corrAddress2 = Str(".//corrAddress2")
    corrAddress3 = Str(".//corrAddress3")
    corrAddress4 = Str(".//corrAddress4")

    def combine_func(self, obj):
        out = str()
        for i in range(1, 5):
            out += obj.get(f"corrAddress{i}", "") + "\n"
        return out.strip()


class AssignmentSchema(Schema):
    id = Str(".//id")
    conveyance_text = Str(
        ".//conveyanceText",
        formatter=lambda s: s.replace("(SEE DOCUMENT FOR DETAILS).", "").strip(),
    )
    date_produced = Date(".//dateProduced")
    corr_name = Str(".//corrName")
    corr_address = CorrespondentAddressField()
    assignment_record_has_images = f.Bool(".//assignmentRecordHasImages", true_value="Y")
    page_count = f.Int(".//pageCount")
    transaction_date = Date(".//transactionDate")
    last_update_date = Date(".//lastUpdateDate")
    recorded_date = Date(".//recordedDate")
    properties = PropertySchema()
    assignors = AssignorSchema()
    assignees = AssigneeSchema()


class AssignmentPageSchema(Schema):
    num_found = f.Int(".//response/@numFound")
    docs = f.List(AssignmentSchema, ".//response/doc")

    def pre_load(self, obj):
        for e in obj.find(".//result").iter():
            if "name" in e.attrib:
                e.tag = e.attrib["name"]
                del e.attrib["name"]
        return obj



import datetime
import typing as tp
from pathlib import Path

from dateutil.parser import parse as parse_dt

from .model import PtabDocumentPage
from .schema import PtabDecisionPageSchema
from .schema import PtabDocumentPageSchema
from .schema import PtabProceedingPageSchema
from .session import session


def convert_date(obj):
    if isinstance(obj, datetime.datetime):
        return obj.date()
    elif isinstance(obj, str):
        return parse_dt(obj).date()
    elif isinstance(obj, tuple):
        return (convert_date(o) for o in obj)


def update_query_with_date(obj, name, query_dict):
    if not obj:
        return query_dict
    obj = convert_date(obj)
    if isinstance(obj, datetime.date):
        query_dict[f"{name}Date"] = obj.isoformat()
    elif isinstance(obj, tuple):
        query_dict[f"{name}FromDate"] = obj[0].isoformat()
        query_dict[f"{name}ToDate"] = obj[1].isoformat()
    return query_dict


parameter_map = {
    "document_title": "documentTitleText",
    "style_name": "styleNameText",
    "filing_party_category": "filingPartyCategory",
    "patent_owner": "patentOwnerName",
    "party_name": "partyName",
    "document_number": "documentIdentifier",
    "document_type": "documentTypeName",
    "document_name": "documentName",
    "proceeding_number": "proceedingNumber",
    "proceeding_type": "proceedingTypeCategory",
    "application_number": "applicationNumberText",
    "start": "recordStartNumber",
    "rows": "recordTotalQuantity",
    "sort": "sortOrderCategory",
    "technology_center_number": "technologyCenterNumber",
    "inventor_name": "inventorName",
    "patent_number": "patentNumber",
    "proceeding_status": "proceedingStatusCategory",
    "subproceeding_type": "subproceedingTypeCategory",
    "decision_type": "decisionTypeCategory",
    "subdecision_type": "subdecisionTypeCategory",
    "text_search": "searchText",
}


class PtabApi:
    @classmethod
    async def get_documents(
        cls,
        document_title: tp.Optional[str] = None,
        document_filing_date: tp.Union[None, datetime.date, tuple[datetime.date], tuple[str]] = None,
        style_name: tp.Optional[str] = None,
        filing_party_category: tp.Optional[str] = None,
        patent_owner_name: tp.Optional[str] = None,
        party_name: tp.Optional[str] = None,
        document_number: tp.Optional[str] = None,
        document_type: tp.Optional[str] = None,
        document_name: tp.Optional[str] = None,
        document_identifier: tp.Optional[str] = None,
        proceeding_number: tp.Optional[str] = None,
        proceeding_type: tp.Optional[str] = None,
        application_number: tp.Optional[str] = None,
        start: tp.Optional[int] = None,
        rows: tp.Optional[int] = None,
        sort: tp.Union[str, tp.List[str], None] = None,
    ) -> PtabDocumentPage:

        query_dict = {
            "documentTitleText": document_title,
            "styleNameText": style_name,
            "filingPartyCategory": filing_party_category,
            "patentOwnerName": patent_owner_name,
            "partyName": party_name,
            "documentNumber": document_number,
            "documentTypeName": document_type,
            "documentName": document_name,
            "documentIdentifier": document_identifier,
            "proceedingNumber": proceeding_number,
            "proceedingTypeCategory": proceeding_type,
            "applicationNumberText": application_number,
            "recordStartNumber": start,
            "recordTotalQuantity": rows,
            "sortOrderCategory": sort,
        }
        query_dict = {k: v for k, v in query_dict.items() if v is not None}

        query_dict = update_query_with_date(document_filing_date, "documentFiling", query_dict)

        response = await session.get(url="https://developer.uspto.gov/ptab-api/documents", params=query_dict)
        response.raise_for_status()
        return PtabDocumentPageSchema().load(response.json())

    @classmethod
    async def download_document(cls, document_identifier: str, output_path):
        output_path = Path(output_path)
        if output_path.exists():
            return output_path
        response = await session.stream(
            "GET",
            url=f"https://developer.uspto.gov/ptab-api/documents/{document_identifier}/download",
        )
        response.raise_for_status()
        with open(output_path, "wb") as f:
            async for chunk in response.aiter_content(chunk_size=8192):
                f.write(chunk)
        return output_path

    @classmethod
    async def get_proceedings(
        cls,
        proceeding_number: tp.Optional[str] = None,
        technology_center_number: tp.Optional[str] = None,
        patent_owner: tp.Optional[str] = None,
        party_name: tp.Optional[str] = None,
        inventor_name: tp.Optional[str] = None,
        patent_number: tp.Optional[str] = None,
        declaration_date: tp.Union[None, datetime.date, tuple[datetime.date], tuple[str]] = None,
        style_name: tp.Optional[str] = None,
        application_number: tp.Optional[str] = None,
        proceeding_status: tp.Optional[str] = None,
        proceeding_type: tp.Optional[str] = None,
        subproceeding_type: tp.Optional[str] = None,
        institution_date: tp.Union[None, datetime.date, tuple[datetime.date], tuple[str]] = None,
        proceeding_filing_date: tp.Union[None, datetime.date, tuple[datetime.date], tuple[str]] = None,
        accorded_filing_date: tp.Union[None, datetime.date, tuple[datetime.date], tuple[str]] = None,
        proceeding_last_modified_date: tp.Union[None, datetime.date, tuple[datetime.date], tuple[str]] = None,
        grant_date: tp.Union[None, datetime.date, tuple[datetime.date], tuple[str]] = None,
        start: tp.Optional[int] = None,
        rows: tp.Optional[int] = None,
        sort: tp.Union[str, tp.List[str], None] = None,
    ) -> PtabDocumentPage:

        query_dict = {
            "proceedingNumber": proceeding_number,
            "technologyCenterNumber": technology_center_number,
            "patentOwnerName": patent_owner,
            "partyName": party_name,
            "inventorName": inventor_name,
            "patentNumber": patent_number,
            "styleNameText": style_name,
            "applicationNumberText": application_number,
            "proceedingStatusCategory": proceeding_status,
            "proceedingTypeCategory": proceeding_type,
            "subproceedingTypeCategory": subproceeding_type,
            "recordStartNumber": start,
            "recordTotalQuantity": rows,
            "sortOrderCategory": sort,
        }
        query_dict = {k: v for k, v in query_dict.items() if v is not None}
        query_dict = update_query_with_date(declaration_date, "declaration", query_dict)
        query_dict = update_query_with_date(institution_date, "institution", query_dict)
        query_dict = update_query_with_date(proceeding_filing_date, "proceedingFiling", query_dict)
        query_dict = update_query_with_date(accorded_filing_date, "accordedFiling", query_dict)
        query_dict = update_query_with_date(proceeding_last_modified_date, "proceedingLastModified", query_dict)
        query_dict = update_query_with_date(grant_date, "grant", query_dict)

        response = await session.get(url="https://developer.uspto.gov/ptab-api/proceedings", params=query_dict)
        response.raise_for_status()
        return PtabProceedingPageSchema().load(response.json())

    @classmethod
    async def get_decisions(
        cls,
        proceeding_number: tp.Optional[str] = None,
        decision_type: tp.Optional[str] = None,
        subdecision_type: tp.Optional[str] = None,
        document_name: tp.Optional[str] = None,
        decision_date: tp.Union[None, datetime.date, tuple[datetime.date], tuple[str]] = None,
        style_name: tp.Optional[str] = None,
        proceeding_type: tp.Optional[str] = None,
        subproceeding_type: tp.Optional[str] = None,
        technology_center_number: tp.Optional[str] = None,
        document_number: tp.Optional[str] = None,
        patent_owner: tp.Optional[str] = None,
        party_name: tp.Optional[str] = None,
        grant_date: tp.Union[None, datetime.date, tuple[datetime.date], tuple[str]] = None,
        patent_number: tp.Optional[str] = None,
        application_number: tp.Optional[str] = None,
        text_search: tp.Optional[str] = None,
        start: tp.Optional[int] = None,
        rows: tp.Optional[int] = None,
        sort: tp.Union[str, tp.List[str], None] = None,
    ) -> PtabDocumentPage:

        query_dict = {
            "proceedingNumber": proceeding_number,
            "decisionTypeCategory": decision_type,
            "subdecisionTypeCategory": subdecision_type,
            "documentName": document_name,
            "styleNameText": style_name,
            "proceedingTypeCategory": proceeding_type,
            "subproceedingTypeCategory": subproceeding_type,
            "technologyCenterNumber": technology_center_number,
            "documentIdentifier": document_number,
            "patentOwnerName": patent_owner,
            "partyName": party_name,
            "patentNumber": patent_number,
            "applicationNumberText": application_number,
            "searchText": text_search,
            "recordStartNumber": start,
            "recordTotalQuantity": rows,
            "sortOrderCategory": sort,
        }
        query_dict = {k: v for k, v in query_dict.items() if v is not None}
        query_dict = update_query_with_date(decision_date, "decision", query_dict)
        query_dict = update_query_with_date(grant_date, "grant", query_dict)

        response = await session.get(url="https://developer.uspto.gov/ptab-api/decisions", params=query_dict)
        response.raise_for_status()
        return PtabDecisionPageSchema().load(response.json())



from patent_client.util import Manager
from yankee.data import ListCollection

from .api import LegalAsyncApi
from .model import LegalEvent
from .schema import LegalSchema


class LegalManager(Manager):
    __schema__ = LegalSchema

    async def aget(self, doc_number, doc_type="publication", format="docdb") -> ListCollection[LegalEvent]:
        return self.__schema__.load(await LegalAsyncApi.get_legal(doc_number, doc_type, format)).events



import datetime

from patent_client.epo.ops.legal import national_codes


def stub_date():
    return datetime.datetime(2022, 8, 1).date()


def test_legal_codes():
    dt = national_codes.current_date
    national_codes.current_date = stub_date
    national_codes.current_date = dt



import datetime

import pytest

from .model import Assignment


class TestAssignment:
    def test_get_assignment(self):
        a = Assignment.objects.get("18247-405")
        assert a.id == "18247-405"
        assert a.conveyance_text == "NUNC PRO TUNC ASSIGNMENT"
        assert a.last_update_date <= datetime.date.fromisoformat("2008-01-22")
        assert a.transaction_date.isoformat() == "2019-07-11"
        assert a.recorded_date.isoformat() == "2006-09-14"
        assert a.corr_name == "JEFFREY H. INGERMAN"
        assert (
            a.corr_address
            == "FISH & NEAVE IP GROUP, ROPES & GRAY LLP\n1251 AVENUE OF THE AMERICAS C3\nNEW YORK, NY 10020-1105"
        )
        assert len(a.assignors) == 1
        assert a.assignors[0].name == "REALTIME DATA COMPRESSION SYSTEMS, INC."
        assert a.assignors[0].ex_date.isoformat() == "2006-09-14"
        assert len(a.assignees) == 1
        assert a.assignees[0].name == "REALTIME DATA LLC"
        assert a.assignees[0].address == "15 WEST 36TH STREET\nNEW YORK, NEW YORK 10018"
        assert len(a.properties) == 5
        assert a.properties[0].invention_title == "SYSTEM AND METHODS FOR ACCELERATED DATA STORAGE AND RETRIEVAL"
        assert a.properties[0].appl_id == "10628795"
        assert a.properties[0].filing_date.isoformat() == "2003-07-28"
        assert a.properties[0].intl_publ_date == None
        assert a.properties[0].intl_reg_num == None
        assert a.properties[0].inventors == "James J. Fallon"
        assert a.properties[0].issue_date.isoformat() == "2006-10-31"
        assert a.properties[0].pat_num == "7130913"
        assert a.properties[0].pct_num == None
        assert a.properties[0].publ_date.isoformat() == "2004-04-15"
        assert a.properties[0].publ_num == "20040073746"

    def test_fetch_assignments_by_assignee(self):
        assignments = Assignment.objects.filter(assignee="US Well Services")
        assert len(assignments) >= 22
        assert assignments[5].assignors[0].name == "OEHRING, JARED"
        assignments = Assignment.objects.filter(assignee="LogicBlox")
        assert len(assignments) >= 9

    def test_fetch_assignments_by_patent(self):
        assignments = Assignment.objects.filter(patent_number="8,789,601")
        assert len(assignments) >= 1
        assert "48041-605" in [a.id for a in assignments]

    def test_fetch_assignments_by_application(self):
        assignments = Assignment.objects.filter(appl_id="14/190,982")
        assert len(assignments) >= 1

    def test_fetch_assignee_with_greater_than_500_assignments(self):
        assignments = Assignment.objects.filter(assignee="Borealis")
        assert len(assignments) >= 1268

    def test_get_assignment_image(self):
        assignments = Assignment.objects.filter(patent_number=6095661)
        assignment = assignments[0]
        assert assignment.image_url == "http://legacy-assignments.uspto.gov/assignments/assignment-pat-038505-0128.pdf"

    def test_slice_assignments(self):
        assignments = Assignment.objects.filter(assignee="US Well Services")
        assignment_list1 = [assignment.id for assignment in assignments[0:5]]
        assert len(assignment_list1) == 5

        assignment_list2 = [assignment.id for assignment in assignments[:5]]
        assert len(assignment_list2) == 5

        assignment_list3 = [assignment.id for assignment in assignments[-5:]]
        assert len(assignment_list3) == 5

    def test_iterate_assignments(self):
        assignments = Assignment.objects.filter(assignee="US Well Services")
        assignment_list = [assignment.id for assignment in assignments]
        assert len(assignment_list) == len(assignments)

    @pytest.mark.skip("Lookup api does not support multiple inputs")
    def test_can_fetch_multiple(self):
        assignments = Assignment.objects.filter(appl_id=["13089872", "15216946"])
        assert assignments.count() == 5


class TestAssignmentBugs:
    def test_id_43433_231(self):
        assignment = Assignment.objects.get("43433-231")
        assert assignment.properties[0].filing_date == datetime.date(2016, 10, 25)

    def test_bug_scidrill(self):
        assignments = Assignment.objects.filter(assignee="Scientific Drilling")
        assert len(assignments) >= 61


class TestAssignmentAnalysis:
    def test_get_apps_from_assignments(self):
        app_ids = (
            Assignment.objects.filter(assignee="US Well Services")
            .explode("properties", unpack=True, prefix=False)
            .values_list("appl_id", flat=True)
            .to_list()
        )
        assert len(app_ids) > 10


class TestAssignmentAsync:
    @pytest.mark.asyncio
    async def test_get_assignment(self):
        a = await Assignment.objects.aget("18247-405")
        assert a.id == "18247-405"
        assert a.conveyance_text == "NUNC PRO TUNC ASSIGNMENT"
        assert a.last_update_date <= datetime.date.fromisoformat("2008-01-22")
        assert a.transaction_date.isoformat() == "2019-07-11"
        assert a.recorded_date.isoformat() == "2006-09-14"
        assert a.corr_name == "JEFFREY H. INGERMAN"
        assert (
            a.corr_address
            == "FISH & NEAVE IP GROUP, ROPES & GRAY LLP\n1251 AVENUE OF THE AMERICAS C3\nNEW YORK, NY 10020-1105"
        )
        assert len(a.assignors) == 1
        assert a.assignors[0].name == "REALTIME DATA COMPRESSION SYSTEMS, INC."
        assert a.assignors[0].ex_date.isoformat() == "2006-09-14"
        assert len(a.assignees) == 1
        assert a.assignees[0].name == "REALTIME DATA LLC"
        assert a.assignees[0].address == "15 WEST 36TH STREET\nNEW YORK, NEW YORK 10018"
        assert len(a.properties) == 5
        assert a.properties[0].invention_title == "SYSTEM AND METHODS FOR ACCELERATED DATA STORAGE AND RETRIEVAL"
        assert a.properties[0].appl_id == "10628795"
        assert a.properties[0].filing_date.isoformat() == "2003-07-28"
        assert a.properties[0].intl_publ_date == None
        assert a.properties[0].intl_reg_num == None
        assert a.properties[0].inventors == "James J. Fallon"
        assert a.properties[0].issue_date.isoformat() == "2006-10-31"
        assert a.properties[0].pat_num == "7130913"
        assert a.properties[0].pct_num == None
        assert a.properties[0].publ_date.isoformat() == "2004-04-15"
        assert a.properties[0].publ_num == "20040073746"

    @pytest.mark.asyncio
    async def test_fetch_assignments_by_assignee(self):
        assignments = Assignment.objects.filter(assignee="US Well Services")
        assert await assignments.alen() >= 22
        assert assignments[5].assignors[0].name == "OEHRING, JARED"
        assignments = Assignment.objects.filter(assignee="LogicBlox")
        assert await assignments.alen() >= 9

    @pytest.mark.asyncio
    async def test_fetch_assignments_by_patent(self):
        assignments = Assignment.objects.filter(patent_number="8,789,601")
        assert await assignments.alen() >= 1
        assert "48041-605" in [a.id async for a in assignments]

    @pytest.mark.asyncio
    async def test_fetch_assignments_by_application(self):
        assignments = Assignment.objects.filter(appl_id="14/190,982")
        assert await assignments.alen() >= 1

    @pytest.mark.asyncio
    async def test_fetch_assignee_with_greater_than_500_assignments(self):
        assignments = Assignment.objects.filter(assignee="Borealis")
        assert await assignments.alen() >= 1268

    @pytest.mark.asyncio
    async def test_get_assignment_image(self):
        assignments = Assignment.objects.filter(patent_number=6095661)
        assignment = await assignments.afirst()
        assert assignment.image_url == "http://legacy-assignments.uspto.gov/assignments/assignment-pat-038505-0128.pdf"

    @pytest.mark.asyncio
    async def test_slice_assignments(self):
        assignments = Assignment.objects.filter(assignee="US Well Services")
        assignment_list1 = [assignment.id async for assignment in assignments[0:5]]
        assert len(assignment_list1) == 5

        assignment_list2 = [assignment.id async for assignment in assignments[:5]]
        assert len(assignment_list2) == 5

        assignment_list3 = [assignment.id async for assignment in assignments[-5:]]
        assert len(assignment_list3) == 5

    @pytest.mark.asyncio
    async def test_iterate_assignments(self):
        assignments = Assignment.objects.filter(assignee="US Well Services")
        assignment_list = [assignment.id async for assignment in assignments]
        assert len(assignment_list) == await assignments.alen()



import datetime
from pathlib import Path

from .schema import AssignmentPageSchema

test_dir = Path(__file__).parent / "test"


def test_assignment_1():
    xml_file = test_dir / "assignment_1.xml"
    parser = AssignmentPageSchema()
    page = parser.load(xml_file.read_text())
    a = page.docs[0]
    assert a.id == "18247-405"
    assert a.conveyance_text == "NUNC PRO TUNC ASSIGNMENT"
    assert a.last_update_date <= datetime.date.fromisoformat("2008-01-22")
    assert a.transaction_date.isoformat() == "2019-07-11"
    assert a.recorded_date.isoformat() == "2006-09-14"
    assert a.corr_name == "JEFFREY H. INGERMAN"
    assert (
        a.corr_address
        == "FISH & NEAVE IP GROUP, ROPES & GRAY LLP\n1251 AVENUE OF THE AMERICAS C3\nNEW YORK, NY 10020-1105"
    )
    assert len(a.assignors) == 1
    assert a.assignors[0].name == "REALTIME DATA COMPRESSION SYSTEMS, INC."
    assert a.assignors[0].ex_date.isoformat() == "2006-09-14"
    assert len(a.assignees) == 1
    assert a.assignees[0].name == "REALTIME DATA LLC"
    assert a.assignees[0].address == "15 WEST 36TH STREET\nNEW YORK, NEW YORK 10018"
    assert len(a.properties) == 5
    assert a.properties[0].invention_title == "SYSTEM AND METHODS FOR ACCELERATED DATA STORAGE AND RETRIEVAL"
    assert a.properties[0].appl_id == "10628795"
    assert a.properties[0].filing_date.isoformat() == "2003-07-28"
    assert a.properties[0].intl_publ_date == None
    assert a.properties[0].intl_reg_num == None
    assert a.properties[0].inventors == "James J. Fallon"
    assert a.properties[0].issue_date.isoformat() == "2006-10-31"
    assert a.properties[0].pat_num == "7130913"
    assert a.properties[0].pct_num == None
    assert a.properties[0].publ_date.isoformat() == "2004-04-15"
    assert a.properties[0].publ_num == "20040073746"


def test_assignment_2():
    xml_file = test_dir / "assignment_2.xml"
    parser = AssignmentPageSchema()
    page = parser.load(xml_file.read_text())
    a = page.docs[0]
    assert len(a.assignors) == 5


def test_assignment_3():
    xml_file = test_dir / "assignment_3.xml"
    parser = AssignmentPageSchema()
    page = parser.load(xml_file.read_text())
    assert page.num_found == 9
    assert len(page.docs) == 9



import importlib
import typing
from dataclasses import dataclass
from dataclasses import fields

from yankee.data.util import DataConversion
from yankee.util import is_valid

ManagerType = typing.TypeVar("ManagerType")


class ModelMeta(type):
    """Metaclass for automatically appending the .objects attriburte to Model
    classes"""

    def __new__(cls, name, bases, dct):
        klass = super().__new__(cls, name, bases, dct)
        return klass

    @property
    def objects(cls):
        if cls.__manager__ is None:
            return None
        if not isinstance(cls.__manager__, str):
            return cls.__manager__()
        obj_module, obj_class = cls.__manager__.rsplit(".", 1)
        return getattr(importlib.import_module(obj_module), obj_class)()


class ModelABC(object):
    __manager__ = None


@dataclass
class Model(ModelABC, DataConversion, metaclass=ModelMeta):
    __exclude_fields__ = list()
    __default_fields__ = False

    # def __init__(self, *args, **kwargs):
    #    try:
    #        return super().__init__(*args, **kwargs)
    #    except TypeError as e:
    #        raise TypeError(f"{e.args[0]}\nargs:{args}\nkwargs:{kwargs}")

    @classmethod
    def fields(cls):
        """Return list of fields"""
        return fields(cls)

    def __iter__(self):
        return self.items()

    def items(self):
        if self.__default_fields__:
            fields = self.__default_fields__
        else:
            fields = sorted(f.name for f in self.fields())
        for f in fields:
            if f in self.__exclude_fields__:
                continue
            value = getattr(self, f, None)
            if is_valid(value):
                yield (f, value)



from patent_client.epo.ops.session import asession


class FamilyAsyncApi:
    @classmethod
    async def get_family(cls, number, doc_type="publication", format="docdb"):
        url = f"http://ops.epo.org/3.2/rest-services/family/{doc_type}/{format}/{number}"
        response = await asession.get(url)
        response.raise_for_status()
        return response.text



import re
from dataclasses import dataclass
from pathlib import Path

base_dir = Path(__file__).parent


@dataclass
class NumberServiceError:
    code: str
    message: str
    kind: str = "MESSAGE"

    def __str__(self):
        return f"{self.code} -- {self.message}"


error_re = re.compile(r"(?P<code>pBRE\d{3,}) (?P<message>.*) (?P<kind>WARNING|ERROR)$")
message_re = re.compile(r"(?P<code>BR(E|W)\d{3,}) (?P<message>.*)$")


def get_errors(string):
    match = error_re.search(string)
    return match.groupdict() if match else None


def get_messages(string):
    match = message_re.search(string)
    return match.groupdict() if match else None


def build_error_dir():

    error_text = (base_dir / "errors.txt").read_text()
    errors = [get_errors(err) for err in error_text.split("\n")]
    message_text = (base_dir / "messages.txt").read_text()
    messages = [get_messages(err) for err in message_text.split("\n")]
    error_objs = [NumberServiceError(**e) for e in errors + messages if e]
    errors = {e.code: e for e in error_objs if e}
    errors["SUCCESS"] = None
    return errors



import importlib
import logging

from .util import resolve

logger = logging.getLogger(__name__)


def get_model(model_name):
    module_name, class_name = model_name.rsplit(".", 1)
    return getattr(importlib.import_module(module_name), class_name)


class OneToOne:
    def __init__(self, related_class_name, attribute=None, **mapping):
        self.many = False
        self.related_class_name = related_class_name
        self.attribute = attribute
        self.mapping = mapping

    def __call__(self):
        module_name, class_name = self.related_class_name.rsplit(".", 1)
        related_class = getattr(importlib.import_module(module_name), class_name)
        filter_obj = {k: getattr(self, v) for (k, v) in self.mapping.items()}
        return resolve(related_class.objects.get(**filter_obj), self.attribute)


class OneToMany:
    def __init__(self, related_class_name, **mapping):
        self.many = False
        self.related_class_name = related_class_name
        self.mapping = mapping

    def __call__(self):
        module_name, class_name = self.related_class_name.rsplit(".", 1)
        related_class = getattr(importlib.import_module(module_name), class_name)
        filter_obj = {k: getattr(self, v) for (k, v) in self.mapping.items()}
        return related_class.objects.filter(**filter_obj)


def one_to_one(related_class_name, attribute=None, **mapping):
    many = False  # For use in introspection with inspect.getclosurevars
    module_name, class_name = related_class_name.rsplit(".", 1)

    @property
    def get(self):
        klass = getattr(importlib.import_module(module_name), class_name)
        filter_obj = {k: getattr(self, v) for (k, v) in mapping.items()}
        logger.debug(f"Fetching related {klass} using filter {filter_obj}")
        result = resolve(klass.objects.get(**filter_obj), attribute)
        return result

    return get


def one_to_many(related_class_name, **mapping):
    many = True  # For use in introspection with inspect.getclosurevars
    module_name, class_name = related_class_name.rsplit(".", 1)

    @property
    def get(self):
        klass = getattr(importlib.import_module(module_name), class_name)
        filter_obj = {k: getattr(self, v) for (k, v) in mapping.items()}
        return klass.objects.filter(**filter_obj)

    return get


def get_manager(class_name):
    module_name, class_name = class_name.rsplit(".", 1)

    @property
    def objects(self):
        klass = getattr(importlib.import_module(module_name), class_name)
        return klass()

    return objects


def recur_accessor(obj, accessor):
    if "__" not in accessor:
        a = accessor
        rest = None
    else:
        a, rest = accessor.split("__", 1)

    if hasattr(obj, a):
        new_obj = getattr(obj, a)
        if callable(new_obj):
            new_obj = new_obj()
    else:
        try:
            a = int(a)
        except ValueError:
            pass
        try:
            new_obj = obj[a]
        except (KeyError, TypeError, IndexError):
            new_obj = None
    if not rest:
        return new_obj
    else:
        return recur_accessor(new_obj, rest)



from pathlib import Path

import pytest

from .api import FamilyAsyncApi

test_dir = Path(__file__).parent / "fixtures" / "examples"
expected_dir = Path(__file__).parent / "fixtures" / "expected"


@pytest.mark.asyncio
async def test_async_example():
    result = await FamilyAsyncApi.get_family("EP1000000A1")
    expected_file = expected_dir / "example.xml"
    expected_file.write_text(result)
    expected = expected_file.read_text()
    assert expected == result



from patent_client.util import Model
from patent_client.util.base.related import get_model
from yankee.data import ListCollection
from yankee.xml.schema import Schema as XmlSchema


class Schema(XmlSchema):
    class Meta:
        namespaces = {
            "ops": "http://ops.epo.org",
            "epo": "http://www.epo.org/exchange",
            "ft": "http://www.epo.org/fulltext",
        }


class InpadocModel(Model):
    @property
    def biblio(self) -> "patent_client.epo.ops.published.model.InpadocBiblio":
        return get_model("patent_client.epo.ops.published.model.InpadocBiblio").objects.get(self.docdb_number)

    @property
    def images(self) -> "patent_client.epo.ops.published.model.Images":
        return get_model("patent_client.epo.ops.published.model.Images").objects.get(self.docdb_number)

    @property
    def description(self) -> "str":
        return get_model("patent_client.epo.ops.published.model.Description").objects.get(self.docdb_number).description

    @property
    def claims(self) -> "patent_client.epo.ops.published.model.Claims":
        return get_model("patent_client.epo.ops.published.model.Claims").objects.get(self.docdb_number)

    @property
    def legal(self) -> "ListCollection[patent_client.epo.ops.legal.model.LegalEvent]":
        return get_model("patent_client.epo.ops.legal.model.Legal").objects.get(self.docdb_number).events

    @property
    def family(
        self,
    ) -> "ListCollection[patent_client.epo.ops.family.model.FamilyMember]":
        return get_model("patent_client.epo.ops.family.model.Family").objects.get(self.docdb_number).family_members



import re
from itertools import zip_longest

from yankee.data import ListCollection
from yankee.util import AttrDict

from .model import Claim

SPLIT_RE = re.compile(r"^\s*([\d\-\.]+[\)\.]|\.Iadd\.[\d\-\.]+\.|\.\[[\d\-\.]+\.)", flags=re.MULTILINE)
NUMERIC_RE = re.compile(r"\d")

LIMITATION_RE = re.compile(r"(\s*[:;]\s*and|\s*[:;]\s*)", flags=re.IGNORECASE)
NUMBER_RE = re.compile(r"(?P<number>\d+)[\)\.]\s+")
WHITESPACE_RE = re.compile(r"\s+")
CLAIM_INTRO_RE = re.compile(r"^[^\d\.\[]+")

DEPENDENCY_RE = re.compile(r"claims? (?P<number>[\d,or ]+)", flags=re.IGNORECASE)
DEPENDENT_CLAIMS_RE = re.compile(r"(?P<number>\d+)([^\d]|$)")
DEPEND_ALL_RE = re.compile(r"(any of the foregoing claims|any of the previous claims)", flags=re.IGNORECASE)

clean_text = lambda text: WHITESPACE_RE.sub(" ", text).strip()


def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


class ClaimsParser(object):
    def parse(self, claim_text):
        claim_strings = self.split_and_clean_claims(claim_text)
        claim_data = [self.parse_claim_string(string) for string in claim_strings]
        claim_dictionary = {c.number: c for c in claim_data}
        for claim in claim_data:
            for d in claim.depends_on:
                claim_dictionary[d].dependent_claims.append(claim["number"])
        return ListCollection(Claim(**d) for d in claim_data)

    def split_and_clean_claims(self, claim_text):
        claim_text = CLAIM_INTRO_RE.sub("", claim_text)
        claim_strs = [claim.strip() for claim in SPLIT_RE.split(claim_text)]
        # Remove preamble - e.g. "We Claim:"
        while not NUMERIC_RE.search(claim_strs[0]):
            claim_strs.pop(0)
        # Group number and subsequent claim together
        claim_strs = list(grouper(claim_strs, 2))
        # Join strings

        claims = list()
        for claim_str in claim_strs:
            claim_number = claim_str[0]
            if "-" in claim_number:  # Handle cancelled claims (e.g. in reissues)
                claim_number = claim_number.replace(".", "")
                start, end, *_ = re.split(r"[^\d]+", claim_number)
                claim_range = list(str(num) for num in range(int(start), int(end) + 1))
                for num in claim_range:
                    claims.append(num + ". " + claim_str[1])
            else:
                claim_string = " ".join(claim_str)
                claims.append(claim_string)
        return claims

    def parse_claim_string(self, text):
        number = int(NUMBER_RE.search(text).group("number"))
        text = NUMBER_RE.sub("", text)
        return AttrDict.convert(
            {
                "number": number,
                # "text": NUMBER_RE.sub("", text),
                "limitations": [clean_text("".join(lim)) for lim in list(grouper(LIMITATION_RE.split(text), 2, ""))],
                "depends_on": self.parse_dependency(text, number),
                "dependent_claims": list(),
            }
        )

    def parse_dependency(self, text, number):
        dependency = DEPENDENCY_RE.search(text)
        if dependency is not None:
            claims = dependency.groupdict()["number"]
            claim_numbers = [int(m.groupdict()["number"]) for m in DEPENDENT_CLAIMS_RE.finditer(claims)]
            return claim_numbers
        elif DEPEND_ALL_RE.search(text):
            return list(range(1, number))
        else:
            return list()



import csv
import re
from collections import defaultdict
from pathlib import Path


class QueryException(Exception):
    pass


class QueryBuilder:
    def __init__(self):
        input_file = Path(__file__).parent / "gd_input.csv"
        with input_file.open() as f:
            reader = csv.DictReader(f)
            self.input_schema = list(reader)
        for row in self.input_schema:
            row["pattern"] = re.compile(row["pattern"])
        self.pattern_index = defaultdict(dict)
        for row in self.input_schema:
            self.pattern_index[row["office_code"]][row["type_code"]] = row["pattern"]

        self.country_codes = ["US", "CN", "EP", "KR", "JP", "AU"]

    def validate_query(self, query):
        pattern = self.pattern_index[query["office_code"]][query["type_code"]]
        if not pattern.fullmatch(query["doc_number"]):
            raise QueryException(
                f"{query['doc_number']} is not a valid number for office: {query['office_code']}, type: {query['type_code']}!"
            )
        return query

    def build_query(self, *args, **kwargs):
        query = self.get_candidate_query(*args, **kwargs)
        return self.validate_query(query)

    def get_candidate_query(self, *args, **kwargs):
        if args and len(args) > 1:
            raise QueryException("Cannot provide more than one non-keyword argument!")
        elif args:
            arg = args[0]
            if arg[:2] in ["US", "CN", "EP", "KR", "JP"]:
                country = arg[:2]
                number = arg[2:]
                if "type" in kwargs:
                    return {
                        "office_code": country,
                        "type_code": kwargs["type"],
                        "doc_number": number,
                    }
                candidates = [
                    r for r in self.input_schema if r["office_code"] == country and r["pattern"].fullmatch(number)
                ]
                if not candidates:
                    raise QueryException(
                        f"While country was detected as {country}, no number format matched {number}, please check your number"
                    )
                if len(candidates) > 1:
                    raise QueryException(
                        f"Cannot determine number type. Possible types are {[c['type_name'] for c in candidates]}"
                    )
                return {
                    "office_code": country,
                    "doc_number": number,
                    "type_code": candidates[0]["type_code"],
                }
            elif arg[:3] == "PCT":
                return {
                    "office_code": "WIPO",
                    "type_code": "application",
                    "doc_number": arg,
                }
            elif arg[:2] == "WO":
                return {
                    "office_code": "WIPO",
                    "type_code": "publication",
                    "doc_number": arg,
                }
            elif arg[:2] == "AU":
                if "type" in kwargs:
                    return {
                        "office_code": "CASE",
                        "type_code": kwargs["type"],
                        "doc_number": arg,
                    }
                else:
                    raise QueryException(
                        "While country was detected as AU, no type can be inferred. Please pass a 'type' keyword to specify application/publication"
                    )
            else:
                office_code = kwargs.get("office", "US")
                candidates = [
                    r for r in self.input_schema if r["office_code"] == office_code and r["pattern"].fullmatch(arg)
                ]
                if not candidates:
                    raise QueryException(
                        f"While country was detected as {office_code}, no number format matched {arg}, please check your number"
                    )
                if len(candidates) > 1:
                    types = [c["type_name"] for c in candidates]
                    if "Application" in types and office_code == "US":
                        return {
                            "office_code": office_code,
                            "doc_number": arg,
                            "type_code": "application",
                        }
                    raise QueryException(f"Cannot determine number type. Possible types are {types}")
                return {
                    "office_code": office_code,
                    "doc_number": arg,
                    "type_code": candidates[0]["type_code"],
                }

        elif kwargs:
            numbers = {k: v for k, v in kwargs.items() if k in ("publication", "application", "patent")}
            if len(numbers) > 1:
                raise QueryException("You may only pass one keyword from the set of (application, publication, patent)")
            elif not numbers:
                raise QueryException("No number passed! Please pass a valid number")
            key, value = list(numbers.items())[0]
            if value[:2] in ["US", "CN", "EP", "KR", "JP"]:
                return {
                    "office_code": value[:2],
                    "type_code": key,
                    "doc_number": value[2:],
                }
            elif value[:2] == "WO" or value[:3] == "PCT":
                return {"office_code": "WIPO", "type_code": key, "doc_number": value}
            elif value[:2] == "AU":
                return {"office_code": "CASE", "type_code": key, "doc_number": value}
            else:
                office_code = kwargs.get("office", "US")
                return {
                    "office_code": office_code,
                    "type_code": key,
                    "doc_number": value,
                }



import json
from pathlib import Path

import lxml.etree as ET
from patent_client.util.test import compare_dicts

from .schema import LegalSchema

test_dir = Path(__file__).parent / "test"
expected_dir = Path(__file__).parent / "test" / "expected"


def test_example():
    tree = ET.parse(test_dir / "example.xml")
    result = LegalSchema().load(tree)
    expected_file = expected_dir / "example.json"
    # expected_file.write_text(result.to_json(indent=2))
    expected = json.loads(expected_file.read_text())
    compare_dicts(json.loads(result.to_json()), expected)


def test_us_example():
    tree = ET.parse(test_dir / "us_example.xml")
    result = LegalSchema().load(tree)
    expected_file = expected_dir / "us_example.json"
    # expected_file.write_text(result.to_json(indent=2))
    expected = json.loads(expected_file.read_text())
    compare_dicts(json.loads(result.to_json()), expected)



import pytest
from dateutil.parser import parse as parse_dt

from .manager import PatentBiblioManager
from .query import QueryException


@pytest.mark.skip("PPS API currently broken :(")
class QueryTest:
    def test_default_query():
        assert PatentBiblioManager().filter("6103599")._query == '"6103599".PN.'

    def test_plural_default_query():
        assert PatentBiblioManager().filter("6103599", "6103600")._query == '("6103599" "6103600").PN.'

    def test_keyword_query():
        assert PatentBiblioManager().filter(appl_id="6103599")._query == '"6103599".APNR.'

    def test_plural_keyword_query():
        assert PatentBiblioManager().filter(appl_id=("11111111", "11222222"))._query == '("11111111" "11222222").APNR.'

    def test_keyword_date_query():
        assert PatentBiblioManager().filter(app_filing_date="2021-01-01")._query == '@APD="20210101"'
        assert PatentBiblioManager().filter(app_filing_date=parse_dt("2021-01-01"))._query == '@APD="20210101"'
        assert PatentBiblioManager().filter(app_filing_date=parse_dt("2021-01-01").date())._query == '@APD="20210101"'

    def test_keyword_date_range_query():
        assert (
            PatentBiblioManager().filter(app_filing_date="2021-01-01->2021-02-01")._query == "@APD>=20210101<=20210201"
        )
        assert (
            PatentBiblioManager().filter(app_filing_date__range=("2021-01-01", "2021-02-01"))._query
            == "@APD>=20210101<=20210201"
        )
        assert (
            PatentBiblioManager().filter(app_filing_date__range=(parse_dt("2021-01-01"), parse_dt("2021-02-01")))._query
            == "@APD>=20210101<=20210201"
        )

    def test_date_operator_queries():
        assert PatentBiblioManager().filter(app_filing_date__gt="2021-01-01")._query == '@APD>"20210101"'
        assert PatentBiblioManager().filter(app_filing_date__gte="2021-01-01")._query == '@APD>="20210101"'
        assert PatentBiblioManager().filter(app_filing_date__lt="2021-01-01")._query == '@APD<"20210101"'
        assert PatentBiblioManager().filter(app_filing_date__lte="2021-01-01")._query == '@APD<="20210101"'

    def test_error_query():
        with pytest.raises(QueryException):
            PatentBiblioManager().filter(blargh="2021-01-01")._query

    def test_error_date_query():
        with pytest.raises(QueryException):
            q = PatentBiblioManager().filter(app_filing_date="2021-00-01")._query
        with pytest.raises(QueryException):
            q = PatentBiblioManager().filter(app_filing_date=False)._query

    def test_multiple_fields():
        assert (
            PatentBiblioManager().filter("6103599", app_filing_date="2021-01-01")._query
            == '@APD="20210101" AND "6103599".PN.'
        )
        assert (
            PatentBiblioManager().filter("6103599").filter(app_filing_date="2021-01-01")._query
            == '"6103599".PN. AND @APD="20210101"'
        )
        assert (
            PatentBiblioManager().filter(patent_number="6103599", app_filing_date="2021-01-01")._query
            == '@APD="20210101" AND "6103599".PN.'
        )

    def test_raw_query():
        assert PatentBiblioManager().filter(query="example query")._query == "example query"

    def test_raw_query_with_keywords():
        assert (
            PatentBiblioManager().filter(query="some text", patent_number="6103599")._query
            == '"6103599".PN. AND some text'
        )

    def test_default_or():
        assert (
            PatentBiblioManager()
            .option(default_operator="OR")
            .filter(patent_number="6103599", appl_id="1234567")
            ._query
            == '"1234567".APNR. OR "6103599".PN.'
        )
